import sys
from collections import UserList
# from scipy.stats import fisher_exact
from itertools import zip_longest
from math import sqrt
import pandas as pd
from IPython.display import Markdown, display

from ..coords import gene_coords
from ..utils import fisher_test

class GeneList(UserList):

    _highlight_color = '#1876D2'
    highlight_color = _highlight_color

    _markup = ['bold', 'color', 'underline', 'italic']
    markup = _markup

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    @classmethod
    def set_highlight_color(cls, color):
        cls.highlight_color = color


    @classmethod
    def reset_highlight_color(cls):
        cls.highlight_color = cls._highlight_color


    def name(self, name=None):
        """
        Set/get a name for the gene list.
        """
        if name is None:
            if hasattr(self, '_name'):
                return self._name
            else:
                return 'unnamed'
        if not isinstance(name, str):
            raise ValueError('Name must be a string.')
        self._name = name
        return self


    def _distance_prune(self, other, distance, assembly):
        """
        Prune genes that are closer than distance to each other.
        """
        
        ovl = self & other
        coords = gene_coords(ovl, assembly=assembly)
        coords = sorted(coords, key=lambda x: (x[0], (x[1] + x[2]) // 2))
        pruned = []
        last_end = -distance
        for i in range(len(coords)):
            chrom, start, end, name = coords[i]
            if start - last_end < distance:
                pruned.append(name)
            else:
                last_end = end
        if not pruned:
            print('No genes removed', file=sys.stderr)
        else:
            print(f'Removed: {", ".join(sorted(pruned))}', file=sys.stderr)
                    
        return GeneList(sorted(set(self).difference(set(pruned))))


    def fisher(self, other, background, min_dist=None, return_counts=False):
        """
        Fisher's exact test for overlap to and another gene list.

        Parameters
        ----------
        other : 
            Other gene list.
        background : 
            Background gene list.
        min_dist : 
            Minimum distance between genes for distance pruning, by default None
        return_counts : 
            Return contingency table, by default False

        Returns
        -------
        :
            p-value from Fisher's exact test (and contingency table if return_counts is True).
        """
        if min_dist is not None:
            # not implemented yet
            raise NotImplementedError('Distance pruning not implemented yet in fisher method.')

        return fisher_test(self, other, background, min_dist=min_dist, return_counts=return_counts)

    # def fisher(self, other, background, min_dist=(None, None), return_counts=False):

    #     distance, assembly = min_dist
    #     a, b = self, other
    #     if min_dist[0] is not None:
    #         a = self._distance_prune(other, distance, assembly)

    #     M = len(background) 
    #     N = len(background & a) 
    #     n = len(background & b)
    #     x = len(background & a & b)
    #     table = [[  x,           n - x          ],
    #             [ N - x,        M - (n + N) + x]]
    #     if return_counts:
    #         return float(fisher_exact(table, alternative='greater').pvalue), table
    #     return float(fisher_exact(table, alternative='greater').pvalue)    
  

    # TODO: add alias mapping to GeneList
    def download_gene_aliases():
        """
        Download mapping from any alias to the cannonical hgcn name for use in set operations.
        """
        ...


    def _tabulate(self):
        """
        Turn list into square'ish matrix
        """
        n = len(self)
        col_width = max(map(len, self)) + 1
        ncols = min(max(100//col_width, 1+sqrt(n/col_width)), 150//col_width)
        # ncols = min(max(80//col_width, 1+sqrt(n/col_width)), 80//col_width)
        nrows = int(n/ncols) + 1
        rows = []
        for r in range(0, n, nrows):
            rows.append(self[r:r+nrows])
        return rows, col_width
        

    def __repr__(self):
        if not len(self):
            return '< Empty GeneList>'
        rows, col_width = self._tabulate()
        repr = []
        for row in list(zip_longest(*rows, fillvalue='')):
            line = []
            for gene in row:
                line.append(gene.ljust(col_width))
            repr.append(''.join(line))
        return('\n'.join(repr))


    def _repr_html_(self):
        if not len(self):
            return '< Empty GeneList >'
        try:
            rows, col_width = self._tabulate()
            style = 'background: transparent!important; line-height: 10px!important;text-align: left!important;'
            table = [f'<table data-quarto-disable-processing="true" {style}>']
            for row in list(zip_longest(*rows, fillvalue='')):
                table.append(f'<tr style="{style}">')
                for gene in row:
                    td_styles = []                
                    if hasattr(self, '_bold') and gene in self._bold:
                        td_styles.append('font-weight: bold;')
                    if hasattr(self, '_color') and gene in self._color:
                        td_styles.append(f'color:{self.highlight_color};')
                    if hasattr(self, '_underline') and gene in self._underline:
                        td_styles.append('text-decoration: underline;')
                    if hasattr(self, '_italic') and gene in self._italic:
                        td_styles.append('font-style: italic;')
                    td_style = style + ' '.join(td_styles)
                    table.append(f'<td style="{td_style}">{gene}</td>')
                table.append('</tr>')
            table.append('</table>')
            self._strip_styles()
            return '\n'.join(table)
        except Exception as e:
            self._strip_styles()
            raise e


    def expand_amplicon_abbrev(self):

        new_list = []
        for gene_name in self:
            abbrev = gene_name.rsplit('_', 1)[0]
            abbrev = abbrev.replace('-', '_')
            if abbrev in AMPL_ABBREV_MAP.keys():
                new_list.extend(AMPL_ABBREV_MAP[abbrev])
            else:
                new_list.append(gene_name)

        # new_list = []
        # for gene_name in old_list:
        #     if gene_name.startswith('amplicon') and '/' in gene_name:
        #         prefix, *variants = gene_name.split('/')
        #         first_amplicon = re.split(r'[_-]+', prefix, 2)[-1]
        #         new_list.append(first_amplicon)
        #         for var in variants:
        #             ampl_name = first_amplicon[:-1] + var
        #             new_list.append(ampl_name)
        #     else:
        #         new_list.append(gene_name)

        self.data = sorted(set(new_list))
    

    def __str__(self):
        return repr(self)


    def _strip_styles(self):
        for attr in ['_bold', '_color', '_italic', '_underline']:
            if hasattr(self, attr):
                delattr(self, attr)
        return self


    def __lshift__(self, other):
        for style in self.markup:
            if not hasattr(self, f'_{style}'):
                setattr(self, f'_{style}', list(other))            
                break
        # if not hasattr(self, '_bold'):
        #     setattr(self, '_bold', list(other))
        # elif not hasattr(self, '_color'):
        #     setattr(self, '_color', list(other))
        # elif not hasattr(self, '_underline'):
        #     setattr(self, '_underline', list(other))
        # elif not hasattr(self, '_italic'):
        #     setattr(self, '_italic', list(other))
        else:
            self._strip_styles()
            raise ValueError('Do not provide more than two three highlight lists')        
        return self


    def __or__(self, other):
        self._strip_styles()
        other._strip_styles()
        return GeneList(sorted(set(self.data + other.data)))


    def __and__(self, other):
        self._strip_styles()
        other._strip_styles()        
        return GeneList(sorted(set(self.data).intersection(set(other.data))))


    def __xor__(self, other):
        self._strip_styles()
        other._strip_styles()        
        inter = set(self.data).intersection(set(other.data))
        union = set(self.data + other.data)
        return GeneList(sorted(union.difference(inter)))


AMPL_ABBREV_MAP = {    
 'amplicon_chrX_CPXCR1': ['CPXCR1'],
 'amplicon_chrX_CSAG1/2/3': ['CSAG1', 'CSAG2', 'CSAG3'],
 'amplicon_chrX_CT45A1/2/3//6/7/8/9/10': ['CT45A1', 'CT45A2', 'CT45A3', 'CT45A6', 'CT45A7', 'CT45A8', 'CT45A9', 'CT45A10'],
 'amplicon_chrX_CT47A1/2/3/4/5/6/7/8/9/10/11/12/B1': ['CT47A1', 'CT47A2', 'CT47A3', 'CT47A4', 'CT47A5', 'CT47A6', 'CT47A7', 'CT47A8', 'CT47A9', 'CT47A10', 'CT47A11', 'CT47A12', 'CT47B1'],
 'amplicon_chrX_CT55': ['CT55'],
 'amplicon_chrX_CT83': ['CT83'],
 'amplicon_chrX_CTAG1A/1B/2': ['CTAG1A', 'CTAG1B', 'CTAG2'],
 'amplicon_chrX_CXorf49/B': ['CXorf49', 'CXorf49B'],
 'amplicon_chrX_CXorf51A/B': ['CXorf51A', 'CXorf51B'],
 'amplicon_chrX_DDX53': ['DDX53'],
 'amplicon_chrX_DMRTC1/B/FAM236A/B/C/D': ['DMRTC1', 'DMRTC1B', 'FAM236A', 'FAM236B', 'FAM236C', 'FAM236D'],
 'amplicon_chrX_EOLA1/2/HSFX3/4': ['EOLA1', 'EOLA2', 'HSFX3', 'HSFX4'],
 'amplicon_chrX_ETD1/B/ZNF75D': ['ETD1', 'ETD1B', 'ZNF75D'],
 'amplicon_chrX_F8/F8A1/2/3/H2AB1/2/3': ['F8', 'F8A1', 'F8A2', 'F8A3', 'H2AB1', 'H2AB2', 'H2AB3'],
 'amplicon_chrX_FAM156A/B': ['FAM156A', 'FAM156B'],
 'amplicon_chrX_FAM47A/B/C': ['FAM47A', 'FAM47B', 'FAM47C'],
 'amplicon_chrX_G6PD/IKBKG': ['G6PD', 'IKBKG'],
 'amplicon_chrX_GAGE10/1/2A/13/12B/12C/12D/12E/12F/12G/12H/12J': ['GAGE10', 'GAGE1', 'GAGE2A', 'GAGE13', 'GAGE12B', 'GAGE12C', 'GAGE12D', 'GAGE12E', 'GAGE12F', 'GAGE12G', 'GAGE12H', 'GAGE12J'],
 'amplicon_chrX_HSFX1/2': ['HSFX1', 'HSFX2'],
 'amplicon_chrX_IL3RA/P2RY8/SLC25A6': ['IL3RA', 'P2RY8', 'SLC25A6'],
 'amplicon_chrX_MAGEA4': ['MAGEA4'],
 'amplicon_chrX_MAGEA12/A2/A2B/A3/A6': ['MAGEA12', 'MAGEA2', 'MAGEA2B', 'MAGEA3', 'MAGEA6'],
 'amplicon_chrX_MAGEA9/9B': ['MAGEA9', 'MAGEA9B'],
 'amplicon_chrX_MAGEB6': ['MAGEB6'],
 'amplicon_chrX_MAGEC1': ['MAGEC1'],
 'amplicon_chrX_MBTPS2/YY2': ['MBTPS2', 'YY2'],
 'amplicon_chrX_NSDHL': ['NSDHL'],
 'amplicon_chrX_NUDT10/11': ['NUDT10', 'NUDT11'],
 'amplicon_chrX_NXF2/2B/5': ['NXF2', 'NXF2B', 'NXF5'],
 'amplicon_chrX_PABPC1L2A/B': ['PABPC1L2A', 'PABPC1L2B'],
 'amplicon_chrX_PAGE2/2B/5': ['PAGE2', 'PAGE2B', 'PAGE5'],
 'amplicon_chrX_RHOXF2/B': ['RHOXF2', 'RHOXF2B'],
 'amplicon_chrX_SPACA5/B': ['SPACA5', 'SPACA5B'],
 'amplicon_chrX_SPANXA1/A2/N1/N2/N3/N4/N5/B1/C/D': ['SPANXA1', 'SPANXA2', 'SPANXN1', 'SPANXN2', 'SPANXN3', 'SPANXN4', 'SPANXN5', 'SPANXB1', 'SPANXC', 'SPANXD'],
 'amplicon_chrX_SSX1/2/2B/344B/5/7': ['SSX1', 'SSX2', 'SSX2B', 'SS344B', 'SSX5', 'SSX7'],
 'amplicon_chrX_SUPT20HL1/2': ['SUPT20HL1', 'SUPT20HL2'],
 'amplicon_chrX_TCEAL2/3/4/5/6': ['TCEAL2', 'TCEAL3', 'TCEAL4', 'TCEAL5', 'TCEAL6'],
 'amplicon_chrX_TCP11X1/2': ['TCP11X1', 'TCP11X2'],
 'amplicon_chrX_TEX28': ['TEX28'],
 'amplicon_chrX_TMEM185A': ['TMEM185A'],
 'amplicon_chrX_VCX/2/3A/3B': ['VCX', 'VCX2', 'VCX3A', 'VCX3B'],
 'amplicon_chrX_XAGE1A/B': ['XAGE1A', 'XAGE1B'],
 'amplicon_chrX_XAGE3': ['XAGE3'],
 'amplicon_chrX_XAGE5': ['XAGE5'],
}


yaml_format = """
<list_label>:
  description: |
    <free text description of the gene list>
    <free text description of the gene list>
  genes: <gene_name>, <gene_name>, ...
<list_label>:
  description: |
    <free text description of the gene list>
    <free text description of the gene list>
  genes: <gene_name>, <gene_name>, ...
"""

class GeneListCollection(object):

    def __init__(self, yaml_file=None, url:str=None, google_sheet:str=None, tab='Sheet1'):

        if yaml_file is not None:
            import yaml
            with open(yaml_file) as f:
                try:
                    self.data = yaml.safe_load(f)
                except yaml.YAMLError as e:
                    print(f"Yaml must honor this this format:\n\n{yaml_format}\n\n", file=sys.stderr)
                    raise e
            for name in self.data.keys():
                if 'genes' not in self.data[name]:
                    raise ValueError(f'Gene list {name} must have a "genes" entry.')
                if 'description' not in self.data[name]:
                    self.data[name]['description'] = ''
        else:
            assert url or google_sheet, 'Either file/url or google_sheet id must be provided.'

            if url is None:
                url = f'https://docs.google.com/spreadsheets/d/{google_sheet}/gviz/tq?tqx=out:csv&sheet={tab}'

            self.desc = []
            for desc in pd.read_csv(url, header=None, low_memory=False).iloc[0]:
                if str(desc) == 'nan':
                    self.desc.append('')
                else:
                    self.desc.append(desc.replace('\n', ' '))
            self.df = pd.read_csv(url, header=1, low_memory=False)
            self.df = self.df.loc[:, [not x.startswith('Unnamed') for x in self.df.columns]]
            self.names = self.df.columns.tolist()
            self.data = {}
            for name, desc in zip(self.names, self.desc):
                self.data[name] = {}
                sr = self.df[name]
                self.data[name]['genes'] = self.df.loc[~sr.isnull(), name].to_list()
                self.data[name]['description'] = desc

    def all_genes(self):
        gene_names = []
        for list_name in self.data.keys():
            gene_names.extend(self.get(list_name))
        return GeneList(sorted(set(gene_names)))
    

    def get(self, name):
        # sr = self.df[name]
        # sr = self.df.loc[~sr.isnull(), name]
        # # lst = sorted(self.expand_amplicon_abbrev(sr.tolist()))
        # lst = sr.tolist()
        return GeneList(self.data[name]['genes']).name(name)


    def _repr_html_(self):
        records = []
        for name in self.data.keys():
            desc = self.data[name]['description']
            if pd.isnull(desc):
                desc = ''
            records.append((name, desc))
        df = pd.DataFrame.from_records(records, columns=['List label', 'Description'])
        # return df._repr_html_()

        # s = df.style.pipe(make_pretty)
        # s.set_table_styles(
        #     {c: [{'selector': '', 'props': [('text-align', 'left')]}] 
        #          for c in df.columns if is_object_dtype(df[c]) and c != 'strand'},
        #     overwrite=False
        # )
        # display(s)

        s = df.style.set_table_styles(
            {c: [{'selector': '', 'props': [('text-align', 'left')]}] 
                 for c in df.columns},
            overwrite=False
        ).hide(axis="index")
        
        return s._repr_html_()


        # out = ['| label | description |', '|:---|:---|']
        # for name in self.data.keys():
        #     desc = self.data[name]['description']
        #     if pd.isnull(desc):
        #         desc = ''
        #     # out.append(f"- **{(name+':**').ljust(130)} {desc}")
        #     out.append(f"| **{name}** | {desc} |")
            
        # display(Markdown('\n'.join(out)))


    def __repr__(self):
        return ""
  
  
    def __iter__(self):
         yield from self.data.keys()



import os, glob, sys
import re
import pandas as pd
from pandas.api.types import is_object_dtype
import numpy as np
from math import isclose, floor, log10
import matplotlib
import matplotlib.pyplot as plt
from IPython.display import Markdown, display
from pandas.api.types import is_object_dtype
from math import sqrt
from collections.abc import Callable
from typing import Any, TypeVar, List, Tuple, Dict, Union
from itertools import zip_longest
from matplotlib.patches import Rectangle
import seaborn as sns
import requests
from collections.abc import Callable
from typing import Any, TypeVar, List, Tuple, Dict, Union
import warnings
from itertools import chain
import shelve
from collections.abc import Sequence
from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.stats import fisher_exact

from .intervals import *


chrom_lengths = {'hg19': {'chr1': 249250621, 'chr2': 243199373, 'chr3': 198022430, 'chr4': 191154276, 
                            'chr5': 180915260, 'chr6': 171115067, 'chr7': 159138663, 'chr8': 146364022, 
                            'chr9': 141213431, 'chr10': 135534747, 'chr11': 135006516, 'chr12': 133851895,
                            'chr13': 115169878, 'chr14': 107349540, 'chr15': 102531392, 'chr16': 90354753, 
                            'chr17': 81195210, 'chr18': 78077248, 'chr19': 59128983, 'chr20': 63025520, 
                            'chr21': 48129895, 'chr22': 51304566, 'chrX': 155270560, 'chrY': 59373566},
                    'hg38': {'chr1': 248956422, 'chr2': 242193529, 'chr3': 198295559, 'chr4': 190214555, 
                            'chr5': 181538259, 'chr6': 170805979, 'chr7': 159345973, 'chr8': 145138636, 
                            'chr9': 138394717, 'chr10': 133797422, 'chr11': 135086622, 'chr12': 133275309, 
                            'chr13': 114364328, 'chr14': 107043718, 'chr15': 101991189, 'chr16': 90338345, 
                            'chr17': 83257441, 'chr18': 80373285, 'chr19': 58617616, 'chr20': 64444167, 
                            'chr21': 46709983, 'chr22': 50818468, 'chrX': 156040895, 'chrY': 57227415}}    

# TODO: make the centromeres fit each assembly!
centromeres = {
    'chr1':    (121700000, 125100000),
    'chr10':   (38000000, 41600000),
    'chr11':   (51000000, 55800000),
    'chr12':   (33200000, 37800000),
    'chr13':   (16500000, 18900000),
    'chr14':   (16100000, 18200000),
    'chr15':   (17500000, 20500000),
    'chr16':   (35300000, 38400000),
    'chr17':   (22700000, 27400000),
    'chr18':   (15400000, 21500000),
    'chr19':   (24200000, 28100000),
    'chr2':    (91800000, 96000000),
    'chr20':   (25700000, 30400000),
    'chr21':   (10900000, 13000000),
    'chr22':   (13700000, 17400000),
    'chr3':    (87800000, 94000000),
    'chr4':    (48200000, 51800000),
    'chr5':    (46100000, 51400000),
    'chr6':    (58500000, 62600000),
    'chr7':    (58100000, 62100000),
    'chr8':    (43200000, 47200000),
    'chr9':    (42200000, 45500000),
    'chrX':    (58100000, 63800000),
    'chrY':    (10300000, 10400000)}    


def black_white(ax):
    """Returns black for light backgrounds, white for dark backgrounds."""
    if ax is None:
        ax = plt.gca()
    bg_color = ax.get_facecolor()
    # Convert to grayscale to determine brightness
    luminance = matplotlib.colors.rgb_to_hsv(matplotlib.colors.to_rgb(bg_color))[2]
    return 'black' if luminance > 0.5 else '#FDFDFD'

def _chrom_sort_key(chrom):
    return [int(x) if x.isdigit() else x for x in re.split(r'(\d+)', chrom)]

def chrom_sort_key(chrom):
    """
    Function for use as key in sorting chromosomes. Works for both
    Python lists and numpy arrays/pandas series.
    """
    if isinstance(chrom, (list, tuple)):
        return chrom_sort_key(chrom[0])
    elif isinstance(chrom, Sequence):
        return [_chrom_sort_key(x) for x in chrom]
    else:
        return _chrom_sort_key(chrom)

def dummy_data():
    df_list = []
    for chrom, length in chrom_lengths['hg38'].items():
        sample = length // 500_000
        Fs = length
        f = np.linspace(5, 30, sample)
        x = np.linspace(0, Fs, sample)
        y = np.sin(2* np.pi * f * x / Fs)
        chrom = [chrom]*sample
        df = pd.DataFrame(dict(x=x, y=y, chrom=chrom)).sort_values('x')
        df['start'] = df.x.shift()
        df['end'] = df.x
        df = df.iloc[1:-1, :]
        df = stairs(df)
        df_list.append(df)
    df = pd.concat(df_list)
    return df


def dummy_segments():
    import random
    segments = []
    for chrom, chrom_len in chrom_lengths['hg38'].items():
        p = sorted([random.randint(0, chrom_len) for _ in range(int(chrom_len//10_000_000))])
        segments.extend([(chrom, *t) for t in zip(p[0::2], p[1::2])])
    return segments
    
use_cache = True
verbose_retrieval = False

CACHE_DIR = os.path.join(os.path.dirname(__file__), 'cache')

class cache_disabled():
    def __enter__(self):
        global use_cache
        use_cache = False

    def __exit__(self, type, value, traceback):
        global disable_cache
        use_cache = True

def fisher_test(one, other, background, min_dist=None, return_counts=False):

    a, b = one, other    

    not_in_background = a ^ (a & background)
    if not_in_background:
        print(f'Removed {len(not_in_background)} genes not in background set', file=sys.stderr)
        print(not_in_background)
    if min_dist is not None:
        try:
            a = one._distance_prune(other, *min_dist)
        except AttributeError as e:
            print('Distance pruning only works for GeneList objects.')
            raise e
        
    M = len(background) 
    N = len(background & a) 
    n = len(background & b)
    x = len(background & a & b)
    table = [[  x,           n - x          ],
            [ N - x,        M - (n + N) + x]]
    if return_counts:
        return float(fisher_exact(table, alternative='greater').pvalue), table
    return float(fisher_exact(table, alternative='greater').pvalue)  

def shelve_it() -> Callable:
    """
    Decorator to cache the result of a function in a shelve file.

    Parameters
    ----------
    file_name : 
        Path to the shelve file
    """
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)
    def decorator(func):
        if not use_cache:
            return func
        def new_func(*args, **kwargs):
            with shelve.open(os.path.join(CACHE_DIR, func.__name__)) as cache:
                key = '*'.join(map(str, args)) + '///' + '**'.join([f'{k}={v}' for k, v in kwargs.items()])
                if key not in cache:
                    if verbose_retrieval:
                        print(f' {func.__name__} retrieving ...', file=sys.stderr)
                    cache[key] = func(*args, **kwargs)
                return cache[key]
        return new_func

    return decorator


def clear_cache(func_name=None):
    """
    Clear the cache of a shelve file.
    """
    if not os.path.exists(CACHE_DIR):
        return
    if func_name is None:
        for file in glob.glob(f'{CACHE_DIR}/*.db'):
            os.remove(file)
    else:
        os.remove(f'{CACHE_DIR}/{func_name}.db')


def _horizon(row, i, cut):
    """
    Compute the values for the three 
    positive and negative intervals.
    """
    val = getattr(row, i)

    if np.isnan(val):
        for i in range(8):
            yield 0
        # for nan color
        yield cut
    else:
        if val < 0:
            for i in range(4):
                yield 0

        val = abs(val)
        for i in range(3):
            yield min(cut, val)
            val = max(0, val-cut)
        yield int(not isclose(val, 0, abs_tol=1e-8)) * cut

        if val >= 0:
            for i in range(4):
                yield 0

        # for nan color
        yield 0


def horizon(df, y=None, ax=None,
                cut=None, # float, takes precedence over quantile_span
                quantile_span = None,
                x='start',
                beginzero=True, 
                offset=0,
                height=None,
                palette='iker',
                colors = None,
                **kwargs):
                # colours = ['#314E9F', '#36AAA8', '#D7E2D4'] + ['midnightblue'] + \
                #           ['#F5DE90', '#F5DE90', '#A51023'] + ['darkred'] + ['whitesmoke']):
                # colors = sns.color_palette("Blues", 3) + ['midnightblue'] + \
                #           sns.color_palette("Reds", 3) + ['darkred'] + ['lightgrey']):

    """
    Horizon bar plot made allowing multiple chromosomes and multiple samples.
    """
    
    palettes = dict(iker = ['#CCE2DF', '#59A9A8', '#374E9B', 'midnightblue',
                            '#F2DE9A', '#DA8630', '#972428', 'darkred',
                             '#D3D3D3'],
                    bluered = sns.color_palette("Blues", 3) + ['midnightblue'] + \
                              sns.color_palette("Reds", 3) + ['darkred'] + \
                              ['lightgrey'],
                          )                          

    if colors is None:
        colors = palettes[palette.replace('_r', '')]

    if palette.endswith('_r'):
        colors = colors[3:6][::-1] + colors[0:3][::-1] + colors[6:]

    # set cut if not set
    if cut is None:
        cut = np.max([np.max(df[y]), np.max(-df[y])]) / 3
    elif quantile_span:
        cut=max(np.abs(np.nanquantile(df[col], quantile_span[0])), 
                np.abs(np.nanquantile(df[col], quantile_span[1]))) / 3,

    # make the data frame to plot
    row_iter = df.itertuples()
    col_iterators = zip(*(_horizon(row, y, cut) for row in row_iter))
    col_names = ['yp1', 'yp2', 'yp3', 'yp4', 
                 'yn1', 'yn2', 'yn3', 'yn4', 'nan']

    df2 = (df[[y, x]]
            .assign(**dict(zip(col_names, col_iterators)))
            )
    df2 = pd.DataFrame(dict((col, list(chain.from_iterable(zip(df2[col].values, df2[col].values)))) for col in df2))

    # make the plot
    with sns.axes_style("ticks"):

        # ingore UserWarning from seaborn that tight_layout is not applied
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            
            # first y tick
            ytic1 = round(cut/3, -int(floor(log10(abs(cut/3)))))

            scale = 1
            if height is not None:
                ymax = max(df2[col_name].max() for col_name in col_names)
                ymin = max(df2[col_name].min() for col_name in col_names)
                scale = height/(ymax-ymin)

            for col_name, color in zip(col_names, colors):
                ax.fill_between(
                    df2[x], 
                    df2[col_name]*scale+offset, 
                    y2=offset,
                    color=color,
                    linewidth=0,
                    capstyle='butt',
                **kwargs)


def map_interval(chrom, start, end, strand, map_from, map_to, species='homo_sapiens'):
    if chrom.startswith('chr'):
        chrom = chrom[3:]
    start, end = int(start), int(end)    
    api_url = f"http://rest.ensembl.org/map/{species}/{map_from}/{chrom}:{start}..{end}:{strand}/{map_to}"
    params = {'content-type': 'application/json'}
    response = requests.get(api_url, params=params)
    if not response.ok:
        response.raise_for_status()
    return response.json()


class nice:

    def __rlshift__(self, df):
        "Left align columns of params frame: df << nice()"

        def make_pretty(styler):

            def commas(v):
                if type(v) is int:
                    s = str(v)[::-1]
                    return ','.join([s[i:i+3] for i in range(0, len(s), 3)])[::-1]
                else:
                    return v

            return styler.format(commas)

        s = df.style.pipe(make_pretty)
        s.set_table_styles(
            {c: [{'selector': '', 'props': [('text-align', 'left')]}] 
                 for c in df.columns if is_object_dtype(df[c]) and c != 'strand'},
            overwrite=False
        )
        display(s)


def tabulate_genes(words, ncols=None):
    n = len(words)
    col_width = max(map(len, words)) + 1
    if ncols is None:
        ncols = max(100//col_width, 1+sqrt(n/col_width))
    nrows = int(n/ncols) + 1
    rows = []
    for r in range(0, n, nrows):
        rows.append(words[r:r+nrows])
    for row in list(zip_longest(*rows, fillvalue='')):
        line = []
        for gene in row:
            line.append(gene.ljust(col_width))
        print(''.join(line))

def fit_lowess(df, x='x', y='y', frac=0.005, it=0, **kwargs):
    return df.groupby('chrom').apply(lambda _df: _fit_lowess(_df, x=x, y=y, frac=frac, it=it, **kwargs)).reset_index()

def _fit_lowess(df, x='x', y='y', frac=0.005, it=0, **kwargs):
    sorted_df = df.sort_values(by=x)
    filtered = lowess(sorted_df[y], sorted_df[x], is_sorted=True, frac=frac, it=it, **kwargs)
    return pd.DataFrame({x: filtered[:,0], y: filtered[:,1]})

def abline(slope, intercept, ax=None):
    "Add a straight line through the plot"
    if ax is None:
        ax = plt.gca()
    x_vals = np.array(ax.get_xlim())
    y_vals = intercept + slope * x_vals
    ax.plot(x_vals, y_vals, '--', color='grey')

def add_band(x_low, x_high, y_low=None, y_high=None, ax=None, color='gray', linewidth=0, alpha=0.5, zorder=0, **kwargs):
    "Plot a gray block on x interval"
    if ax is None:
        ax = plt.gca()
    if y_low is None:
        y_low, _ = ax.get_ylim()
    if y_high is None:
        _, y_high = ax.get_ylim()
    g = ax.add_patch(Rectangle((x_low, y_low), x_high-x_low, y_high-y_low, 
                 facecolor=color,
                 linewidth=linewidth,
                 alpha=alpha,
                 zorder=zorder,
                 **kwargs))

def stairs(df, start='start', end='end', pos='pos', endtrim=0):
    "Turn a df with start, end into one with pos to plot as stairs"
    df1 = df.copy(deep=True)
    df2 = df.copy(deep=True)
    df1[pos] = df1[start]
    df2[pos] = df2[end] - endtrim
    return pd.concat([df1, df2]).sort_values([start, end])




import numpy as np
from collections import defaultdict
import pandas as pd
from pandas.api.types import is_object_dtype
from collections.abc import Callable, MutableSequence
from typing import Any, TypeVar, List, Tuple, Dict, Union, Iterable

from ..intervals import *
from ..utils import horizon
from ..coords import gene_coords

import math
from math import isclose, floor, log10
from operator import sub
import warnings
from itertools import chain
from statsmodels.nonparametric.smoothers_lowess import lowess
import re

import matplotlib
import matplotlib.axes
import matplotlib.pyplot as plt
import matplotlib.patches
from matplotlib import patches
from matplotlib.artist import Artist
import matplotlib.gridspec as gridspec
from matplotlib.transforms import (Bbox, TransformedBbox,
                                   blended_transform_factory)
from mpl_toolkits.axes_grid1.inset_locator import (BboxConnector,
                                                   BboxConnectorPatch,
                                                   BboxPatch)
import seaborn as sns

from matplotlib.text import OffsetFrom
import textwrap

from ..utils import chrom_lengths, centromeres

from .genome import GenomeIdeogram, _get_chrom_info

class ChromIdeogram(GenomeIdeogram):
    """
    Child class of GenomeIdeogram for plotting single chromosome ideograms.
    """

    def __init__(self, chrom:str, assembly:str, axes_height_inches:int=2, axes_width_inches:int=12,
                 hspace:float=0.3, ylim:tuple=(0, 10), zooms:list=[], zoom_font_size:float=None,
                 zoom_height_ratio=1.0, zoom_effect_color:str='lightgray', zoom_effect_alpha:float=0.3,
                 zoom_height_edgecolor:str='black',
                 wspace:float=0.1, font_size:float=4,
                ):
        """
        Initialize canvas for plotting an ideogram for a single chromosomes.

        Parameters
        ----------
        chrom : 
            Chromosome identifier (E.g. 'chr1', 'chrX').
        assembly : 
            Genome assembly as ucsc identifier (E.g. hg39, rheMac10).
        axes_height_inches : 
            Height of panel for each chromosome ideogram, by default 0.5
        axes_width_inches : 
            Width of panel for longest chromosome ideogram, by default 12
        hspace : 
            Space between additional axes, by default 0
        ylim : 
            Value range on y-axis for placing elements, by default (0, 10)
        font_size : 
            Font size as percentage of panel height (axes_height_inches), 
            by default 4
        zooms : 
            List of chromosome intervals (start, end) to show in seperate detail plots.
        zoom_height_ratio : 
            Height of zoom detail plots relative to height of main plot.
        zoom_font_size : 
            Height of zoom detail plots relative to height of main plot, by default same as font_size
        zoom_effect_color : 
            Face color of zoom effect patches, by default 'lightgray'
        zoom_effect_alpha: 
            Alpha of zoom effect patches, by default 0.3
        zoom_effect_edgecolor: 
            Edge color of zoom effect patches, by default 'black'
        """
        # self.species = species
        self.assembly = assembly
        self.ideogram_base = None
        self.ideogram_height = None
        self.legend_handles = []
        self.zooms = zooms
        self.zoom_axes = []
        self.height_ratios = [1]        
        self.end_padding = 300000
        self.chr_names = [chrom]

        self.zoom_height_ratio = zoom_height_ratio
        self.zoom_effect_color = zoom_effect_color
        self.zoom_effect_alpha = zoom_effect_alpha
        self.zoom_effect_edgecolor = zoom_height_edgecolor

        # self.chr_sizes = [self.chrom_lengths[assembly][chrom] 
        #                   for chrom in self.chr_names]
        self.chr_names = [chrom]
        if self.assembly is not None and assembly in self._chrom_lengths:
            #self.chr_names = [f'chr{x}' for x in list(range(1, 23))+['X', 'Y']]
            self.chr_sizes = [self._chrom_lengths[assembly][c] for c in self.chr_names]
            self.centromeres = self._centromeres
            self.coord_system = {'hg38':'GRCh38', 
                                 'hg19':'GRCh37',
                                 'hg37':'GRCh37',
                                 }[assembly]
        else:
            self.chrom_lengths, self.centromeres = _get_chrom_info(self.assembly)
            self.chr_sizes = [self.chrom_lengths[chrom]]

        self.max_chrom_size = max(self.chr_sizes)
        self.aspect = axes_height_inches / axes_width_inches
        fig_height_inches = axes_height_inches
        point_size = 1/72
        self.ylim = ylim
        self.font_size = font_size/100 * axes_height_inches / point_size 

        if zooms:
            fig_height_inches += (1 + hspace) * axes_height_inches




        if zoom_font_size is None:
            self.zoom_font_size = self.font_size
        else:
            self.zoom_font_size = zoom_font_size/100 * axes_height_inches / point_size 

        fig_width_inches = axes_width_inches
        figsize = (fig_width_inches, fig_height_inches)

        with plt.rc_context(self.d):

            xlim = (-self.end_padding, self.chr_sizes[0]+self.end_padding)
            scaled_y_lim = xlim[0] * self.aspect, xlim[1] * self.aspect

            if zooms:
                self.fig =  plt.figure(figsize=figsize)
                axs = self.fig.subplot_mosaic([
                    ["main" for i in range(len(zooms))],
                    [f"zoom{i}" for i in range(len(zooms))]],
                     width_ratios=[-sub(*tup) for tup in zooms],
                     height_ratios=[1, self.zoom_height_ratio]
                )
                axs["main"].set(xlim=xlim)

                for i in range(len(zooms)):
                    self.zoom_effect(axs[f"zoom{i}"], axs["main"], 
                                     alpha=self.zoom_effect_alpha, 
                                     facecolor=self.zoom_effect_color,
                                     edgecolor=self.zoom_effect_edgecolor)
                    axs[f"zoom{i}"].set(xlim=zooms[i])
                    axs[f"zoom{i}"].set_facecolor((0, 0, 0, 0))
                    # axs[f"zoom{i}"].set_xticks(np.arange(0, self.chr_sizes[0]+1, 10_000_000))
                    # axs[f"zoom{i}"].xaxis.tick_bottom()
                    xtl = axs[f"zoom{i}"].get_xticklabels()
                    xtl[-1] = ""
                    axs[f"zoom{i}"].get_xticklabels(xtl)

                    zoom_scaled_y_lim = self._scaled_y_lim(axs[f"zoom{i}"])
                    axs[f"zoom{i}"].set(ylim=zoom_scaled_y_lim)

                self.ax = axs["main"]
                self.zoom_axes = [axs[f"zoom{i}"] for i in range(len(zooms))]
                
                for i in range(len(self.zoom_axes)):
                    self.zoom_axes[i].sharey(self.zoom_axes[0])
                    self.zoom_axes[i].yaxis.set_ticks_position('none')
                    self.zoom_axes[i].set_yticklabels([])

                    from matplotlib.ticker import (MultipleLocator, AutoMinorLocator)
                    # self.zoom_axes[i].xaxis.set_major_locator(MultipleLocator(1_000_000))
                    self.zoom_axes[i].xaxis.set_minor_locator(MultipleLocator(200_000))
                    x_ticks = self.zoom_axes[i].xaxis.get_major_ticks()
                    x_ticks[-1].label1.set_visible(False)

                plt.subplots_adjust(wspace=wspace)

            else:        
                self.fig, self.ax = plt.subplots(1, 1, figsize=figsize)
            
            if hspace is not None:
                plt.subplots_adjust(hspace=hspace)
    
            plt.minorticks_on()
        
            gs = matplotlib.gridspec.GridSpec(1, 1)
            gs.update(wspace=0, hspace=hspace) 

            self.ax.set_xlim(xlim)
            self.ax.set_ylim(scaled_y_lim)                

            self.ax.spines[['right', 'top', 'left']].set_visible(False)
            x = -self.end_padding * 10 / figsize[1]
            self.ax.set_yticklabels([])
            self.ax.set_xticks(np.arange(0, self.chr_sizes[0]+1, 10_000_000))
            self.ax.xaxis.tick_bottom()
            self.ax.spines['bottom'].set_visible(True)                    
            self.ax.yaxis.set_ticks_position('none')
            self.ax.yaxis.set_visible(False)
            self.ax.set_yticklabels([])

            self.ax_list = [self.ax]
        
            self.chr_axes = dict(zip(self.chr_names, self.ax_list))

    
    def connect_bbox(self, bbox1, bbox2,
                     loc1a, loc2a, loc1b, loc2b,
                     prop_lines, prop_patches=None):
        if prop_patches is None:
            prop_patches = {
                **prop_lines,
                "alpha": prop_lines.get("alpha", 1) * 0.2,
                "clip_on": False,
            }
    
        c1 = BboxConnector(
            bbox1, bbox2, loc1=loc1a, loc2=loc2a, clip_on=False, **prop_lines)
        c2 = BboxConnector(
            bbox1, bbox2, loc1=loc1b, loc2=loc2b, clip_on=False, **prop_lines)
    
        bbox_patch1 = BboxPatch(bbox1, ec='none', fc='none')
        bbox_patch2 = BboxPatch(bbox2, **prop_patches)
    
        p = BboxConnectorPatch(bbox1, bbox2,
                               loc1a=loc1a, loc2a=loc2a, loc1b=loc1b, loc2b=loc2b,
                               clip_on=False,
                               **prop_patches)
    
        return c1, c2, bbox_patch1, bbox_patch2, p
    
    
    def zoom_effect(self, ax1, ax2, **kwargs):
        """
        ax1 : the main Axes
        ax1 : the zoomed Axes
    
        Similar to zoom_effect01.  The xmin & xmax will be taken from the
        ax1.viewLim.
        """
    
        tt = ax1.transScale + (ax1.transLimits + ax2.transAxes)
        trans = blended_transform_factory(ax2.transData, tt)
    
        mybbox1 = ax1.bbox
        mybbox2 = TransformedBbox(ax1.viewLim, trans)
    
        defaults = {'edgecolor': 'black', 'facecolor': 'lightgray', 'alpha': 0.3, 'linewidth': 0.5}
        defaults.update(kwargs)

        # prop_patches = defaults
        # del kwargs['alpha']

        c1, c2, bbox_patch1, bbox_patch2, p = self.connect_bbox(
            mybbox1, mybbox2,
            loc1a=2, loc2a=3, loc1b=1, loc2b=4,
            prop_lines={}, 
            prop_patches=defaults)
    
        ax1.add_patch(bbox_patch1)
        ax2.add_patch(bbox_patch2)
        ax2.add_patch(c1)
        ax2.add_patch(c2)
        ax2.add_patch(p)
    
        return c1, c2, bbox_patch1, bbox_patch2, p


    def add_axes(self, nr_axes=1, height_ratio=1.0, hspace=None):

        new_axes = []
        for _ in range(nr_axes):
        
            gs = self.ax.get_gridspec()
            row = gs.nrows + 1 
            self.height_ratios.append(height_ratio)
            gs = gridspec.GridSpec(row, 1, height_ratios=self.height_ratios,
                                   hspace=hspace)
            for i, ax in enumerate(self.fig.axes):
                ax.set_position(gs[i].get_position(self.fig))
                ax.set_subplotspec(gs[i])
            new_ax = self.fig.add_subplot(gs[row-1], sharex=ax)
            self.fig.set_figheight(
                self.fig.get_figheight() * sum(self.height_ratios) \
                / sum(self.height_ratios[:-1]))
            new_ax.spines[['right', 'top']].set_visible(False)
            new_ax.xaxis.set_visible(False)
            new_axes.append(new_ax)

        new_axes[-1].xaxis.tick_bottom()
        new_axes[-1].xaxis.set_visible(True)

        if nr_axes == 1:
            return new_axes[0]
        else:
            return new_axes






import numpy as np
from collections import defaultdict
import pandas as pd
from pandas.api.types import is_object_dtype
from collections.abc import Callable, MutableSequence
from typing import Any, TypeVar, List, Tuple, Dict, Union, Iterable
from itertools import cycle

from ..intervals import *
from ..utils import horizon, black_white
from ..coords import gene_coords, chromosome_lengths, centromere_coords

import math
from math import isclose, floor, log10
from operator import sub
import warnings
from itertools import chain
from statsmodels.nonparametric.smoothers_lowess import lowess
import requests, re

import matplotlib
import matplotlib.axes
import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.colors
from matplotlib import patches
from matplotlib.artist import Artist
import matplotlib.gridspec as gridspec
from matplotlib.transforms import (Bbox, TransformedBbox,
                                   blended_transform_factory)
from mpl_toolkits.axes_grid1.inset_locator import (BboxConnector,
                                                   BboxConnectorPatch,
                                                   BboxPatch)
import seaborn as sns

from matplotlib.text import OffsetFrom
import textwrap

# from ..utils import chrom_lengths, centromeres

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

class Polygon:
    def __init__(self, points):
        self.points = points

    def get_points(self):
        return self.points

    def nudge_y(self, nudge):
        for point in self.points:
            point.y += nudge

def _get_chrom_info(assembly):
    lengths = {c: l for c, l in chromosome_lengths(assembly)}
    centromeres = {c: (s, e) for c, s, e in centromere_coords(assembly)}
    return lengths, centromeres


# def _get_chrom_info(species):

#     name = species.lower().replace(' ', '_')

#     server = "https://rest.ensembl.org"
#     # ext = "/info/assembly/homo_sapiens?bands=1?"
#     # ext = "/info/assembly/papio_anubis?bands=1?"
#     ext = f"/info/assembly/{name}?bands=1?"
    
#     r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    
#     if not r.ok:
#         r.raise_for_status()
    
#     data = r.json()
#     assembly_versions = data['coord_system_versions']
#     centromeres = {}
#     chrom_lengths = {}
    
#     assembly_coord_system = data['default_coord_system_version']
#     karyotype = data['karyotype']
#     assembly_accession = data['assembly_accession']
        
#     for d in data['top_level_region']:
#         if d['coord_system'] not in ['chromosome', 'primary_assembly']:
#             continue
#         if d['name'] not in karyotype:# .startswith('Un') or d['length'] < 1_000_000:
#             continue
#         if d['name'] == 'MT':# .startswith('Un') or d['length'] < 1_000_000:
#             continue
#         chrom_lengths['chr'+d['name']] = d['length']
    
        
#         if 'bands' in d:
#             for b in d['bands']:
#                 if b['stain'] == 'acen':
#                     centromeres['chr'+b['seq_region_name']] = (b['start'],  b['end'])


#     return assembly_coord_system, chrom_lengths, centromeres




class GenomeIdeogram:
    """
    Class to plot ideograms of chromosomes in a genome assembly.
    """

    d = {'axes.linewidth': 0.8, 'grid.linewidth': 0.64, 
         'lines.linewidth': 0.96, 'lines.markersize': 3.84, 
         'patch.linewidth': 0.64, 'xtick.major.width': 0.8,
         'ytick.major.width': 0.8, 'xtick.minor.width': 0.64, 
         'ytick.minor.width': 0.64, 'xtick.major.size': 3.84, 
         'ytick.major.size': 3.84, 'xtick.minor.size': 2.56, 
         'ytick.minor.size': 2.56, 'font.size': 7.68, 
         'axes.labelsize': 7.68, 'axes.titlesize': 7.68, 
         'xtick.labelsize': 7.04, 'ytick.labelsize': 7.04, 
         'legend.fontsize': 7.04, 'legend.title_fontsize': 7.68}

    _chrom_lengths = {'hg19': {'chr1': 249250621, 'chr2': 243199373, 
                              'chr3': 198022430, 'chr4': 191154276, 
                              'chr5': 180915260, 'chr6': 171115067, 
                              'chr7': 159138663, 'chr8': 146364022, 
                              'chr9': 141213431, 'chr10': 135534747, 
                              'chr11': 135006516, 'chr12': 133851895,
                              'chr13': 115169878, 'chr14': 107349540, 
                              'chr15': 102531392, 'chr16': 90354753, 
                              'chr17': 81195210, 'chr18': 78077248, 
                              'chr19': 59128983, 'chr20': 63025520, 
                              'chr21': 48129895, 'chr22': 51304566, 
                              'chrX': 155270560, 'chrY': 59373566},
                     'hg38': {'chr1': 248956422, 'chr2': 242193529, 
                              'chr3': 198295559, 'chr4': 190214555, 
                              'chr5': 181538259, 'chr6': 170805979, 
                              'chr7': 159345973, 'chr8': 145138636, 
                              'chr9': 138394717, 'chr10': 133797422, 
                              'chr11': 135086622, 'chr12': 133275309, 
                              'chr13': 114364328, 'chr14': 107043718, 
                              'chr15': 101991189, 'chr16': 90338345, 
                              'chr17': 83257441, 'chr18': 80373285, 
                              'chr19': 58617616, 'chr20': 64444167, 
                              'chr21': 46709983, 'chr22': 50818468, 
                              'chrX': 156040895, 'chrY': 57227415}}    

    # TODO: make the centromeres fit each assembly!
    _centromeres = {
        'chr1':    (121700000, 125100000),
        'chr10':   (38000000, 41600000),
        'chr11':   (51000000, 55800000),
        'chr12':   (33200000, 37800000),
        'chr13':   (16500000, 18900000),
        'chr14':   (16100000, 18200000),
        'chr15':   (17500000, 20500000),
        'chr16':   (35300000, 38400000),
        'chr17':   (22700000, 27400000),
        'chr18':   (15400000, 21500000),
        'chr19':   (24200000, 28100000),
        'chr2':    (91800000, 96000000),
        'chr20':   (25700000, 30400000),
        'chr21':   (10900000, 13000000),
        'chr22':   (13700000, 17400000),
        'chr3':    (87800000, 94000000),
        'chr4':    (48200000, 51800000),
        'chr5':    (46100000, 51400000),
        'chr6':    (58500000, 62600000),
        'chr7':    (58100000, 62100000),
        'chr8':    (43200000, 47200000),
        'chr9':    (42200000, 45500000),
        'chrX':    (58100000, 63800000),
        'chrY':    (10300000, 10400000)}                     

    
    def __init__(self, assembly:str, axes_height_inches:float=0.5, axes_width_inches:float=12, 
                 hspace:float=0, ylim:tuple=(0, 10), font_size:float=10, 
                #  species:str='Homo sapiens', 
                 ):
        """
        Initialize canvas for plotting ideograms of chromosomes in a genome
        assembly.

        Parameters
        ----------
        assembly : 
            Genome assembly as ucsc identifier (E.g. hg39, rheMac10).
        axes_height_inches : 
            Height of panel for each chromosome ideogram, by default 0.5
        axes_width_inches : 
            Width of panel for longest chromosome ideogram, by default 12
        hspace : 
            Space between additional axes, by default 0
        ylim : 
            Value range on y-axis for placing elements, by default (0, 10)
        font_size : 
            Font size as percentage of panel height (axes_height_inches), 
            by default 4
        # species : 
        #     Species latin name, by default panel 'Homo sapiens'.
        """
        
        # self.species = species
        self.assembly = assembly
        self.ideogram_base = None
        self.ideogram_height = None
        # self.min_stick_height = min_stick_height
        self.legend_handles = []
        self.zooms = []
        self.zoom_axes = []
        self.zoom_height_ratio = 1.0
        self.zoom_effect_color = 'lightgray'
        self.zoom_effect_alpha = 0.3
        self.end_padding = 300000

        # if self.assembly is not None and assembly in self._chrom_lengths:
        #     self.chrom_lengths = self._chrom_lengths[assembly]
        #     self.chr_names = [f'chr{x}' for x in list(range(1, 23))+['X', 'Y']]
        #     self.chr_sizes = [self.chrom_lengths[c] for c in self.chr_names]
        #     self.centromeres = self._centromeres
        # else:
        #     self.coord_system, self.chrom_lengths, self.centromeres = _get_chrom_info(self.species)
        #     def sort_key(name):
        #         _, num, alp = re.match(r'(chr)?(\d*)(.*)', name).groups()
        #         if not num:
        #             num = '999'
        #         return int(num), alp
                
        #     self.chr_names = sorted(self.chrom_lengths.keys(), key=sort_key)
        #     self.chr_sizes = [self.chrom_lengths[c] for c in self.chr_names]

        self.chrom_lengths, self.centromeres = _get_chrom_info(self.assembly)
        def sort_key(name):
            _, num, alp = re.match(r'(chr)?(\d*)(.*)', name).groups()
            if not num:
                num = '999'
            return int(num), alp
            
        self.chr_names = sorted(self.chrom_lengths.keys(), key=sort_key)
        self.chr_sizes = [self.chrom_lengths[c] for c in self.chr_names]

        # self.chr_names = [f'chr{x}' for x in list(range(1, 23))+['X', 'Y']]
        # self.chr_sizes = [self.chrom_lengths[assembly][chrom] 
        #                   for chrom in self.chr_names]
        self.max_chrom_size = max(self.chr_sizes)
        nr_rows = len(self.chr_names) - 1
        self.aspect = axes_height_inches / axes_width_inches
        axes_width = self.max_chrom_size + 2 * self.end_padding
        axes_height = self.aspect * axes_width
        fig_height_inches = axes_height_inches * (nr_rows-1) + hspace * (nr_rows-1)
        fig_width_inches = axes_width_inches
        figsize = (fig_width_inches, fig_height_inches)
        point_size = 1/72
        self.ylim = ylim
        self.font_size = font_size/100 * axes_height_inches / point_size

        self.zoom_font_size = self.font_size

        self.fig = plt.figure(figsize=figsize)
        plt.subplots_adjust(hspace=0)
        
        with plt.rc_context(self.d):
        
            nr_rows, nr_cols = len(self.chr_names)-2+1, 2
            gs = matplotlib.gridspec.GridSpec(
                nr_rows, 19, height_ratios=[0.4]+[1]*(nr_rows-1))
            gs.update(wspace=0, hspace=hspace) 

            dummy_ax = plt.subplot(gs[0, :])
            xlim = (-self.end_padding, self.max_chrom_size+self.end_padding)
            dummy_ax.set_xlim(xlim)
            scaled_y_lim = xlim[0] * self.aspect, xlim[1] * self.aspect
            # dummy_ax.set_ylim(scaled_y_lim)
            # dummy_ax.plot(1, 1)

            dummy_ax.spines['bottom'].set_visible(False)
            dummy_ax.spines['left'].set_visible(False)
            dummy_ax.spines['right'].set_visible(False)
            dummy_ax.spines['top'].set_visible(True)
            dummy_ax.xaxis.tick_top()
            dummy_ax.xaxis.set_label_position('top') 
            dummy_ax.yaxis.set_ticks_position('none')

            dummy_ax.set_yticklabels([])
        
            ax_list = [plt.subplot(gs[i, :]) for i in range(1, nr_rows-2)]

            ax_list.append(plt.subplot(gs[nr_rows-2, :12]))
            ax_list.append(plt.subplot(gs[nr_rows-1, :12]))
            ax_list.append(plt.subplot(gs[nr_rows-2, 7:]))
            ax_list.append(plt.subplot(gs[nr_rows-1, 7:]))

            self.ax_list = ax_list
            self.chr_axes = dict(zip(self.chr_names, self.ax_list))

            for i in range(len(self.ax_list)):
                chrom = self.chr_names[i]
                ax = ax_list[i]

                if i < len(self.ax_list) - 4:
                    xlim = (-self.end_padding, self.max_chrom_size+self.end_padding)
                    scaled_y_lim = xlim[0] * self.aspect, xlim[1] * self.aspect
                    ax.set_xlim(xlim)
                    ax.set_ylim(scaled_y_lim)
                elif i < len(self.ax_list) - 2:
                    xlim = (-self.end_padding, 12/19*self.max_chrom_size+self.end_padding)
                    scaled_y_lim = xlim[0] * self.aspect, xlim[1] * self.aspect * 19/12
                    ax.set_xlim(xlim)
                    ax.set_ylim(scaled_y_lim) 
                else:
                    xlim = (-self.end_padding, self.chrom_lengths['chrX']+self.end_padding)
                    scaled_y_lim = xlim[0] * self.aspect, xlim[1] * self.aspect * 19/12
                    # \
                    #     * (self.max_chrom_size+2*self.end_padding) \
                    #     / (self.chrom_lengths['hg38']['chrX']+2*self.end_padding)
                    ax.set_xlim(xlim)
                    ax.set_ylim(scaled_y_lim) 
                start, end = 0, self.chr_sizes[i]
                ax.spines[['right', 'top', 'left', 'bottom']].set_visible(False)

                if i in [20, 21]:   
                    x = -2000000 * 10 / figsize[1]
                else:
                    x = -2000000 * 10 / figsize[1]

                ax.text(x, 
                        self.map_y(-sub(*self.ylim), ax)/2, 
                        chrom.replace('chr', ''), 
                        fontsize=7, 
                        verticalalignment='center', 
                        horizontalalignment='right', 
                        weight='bold')
                ax.set_yticklabels([])

                if i == len(self.ax_list)-1:
                    ax.xaxis.tick_bottom()
                    ax.spines['bottom'].set_visible(True)                    
                    ax.yaxis.set_ticks_position('none')
                    ax.yaxis.set_visible(False)
                    ax.set_yticklabels([])
                else:
                    ax.set_xticklabels([])
                    ax.xaxis.set_ticks_position('none')
                    ax.yaxis.set_ticks_position('none')


    def draw_chromosomes(self, base:float=4, height:float=2, 
                         facecolor:str='#EBEAEA', edgecolor:str='black', 
                         linewidth:float=0.7, hide_sex_chrom=False,**kwargs:dict) -> None:
        """
        Draws chromosome ideograms.

        Parameters
        ----------
        base : 
            Placement of ideogram lower edge on y-axis scale, by default 4
        height : 
            Height of ideogram on y-axis scale, by default 2
        facecolor : 
            Ideogram fill color, by default '#EBEAEA'
        edgecolor : 
            Ideogram edge color, by default 'black'
        linewidth : 
            Ideogram edge width, by default 0.7
        hide_sex_chrom :
            If True, hides X and Y chromosomes, by default False
        **kwargs :
            Additional keyword arguments for matplotlib.patches.Rectangle
        """
        self.ideogram_base = base
        self.ideogram_height = height
        
        with plt.rc_context(self.d):

            for i in range(len(self.ax_list)):
                chrom = self.chr_names[i]
                ax = self.ax_list[i]
                
                if hide_sex_chrom and chrom in ['chrX', 'chrY']:
                    ax.set_axis_off()
                    continue

                start, end = 0, self.chr_sizes[i]                

                ideogram_base = self.map_y(base, ax)
                ideogram_height = self.map_y(height, ax)
                # draw centromere
                ymin, ymax = ax.get_ylim()
                if chrom in self.centromeres:
                    cent_start, cent_end = self.centromeres[chrom]
                    xy = [[cent_start, ideogram_base], 
                        [cent_start, ideogram_base+ideogram_height], 
                        [cent_end, ideogram_base], 
                        [cent_end, ideogram_base+ideogram_height]]
                    g = ax.add_patch(patches.Polygon(xy, closed=True, zorder=2, 
                                                    fill=True, color='#777777'))
                else:
                    cent_start, cent_end = 0, 0 

                # draw chrom
                g = ax.add_patch(patches.Rectangle((start, ideogram_base), 
                                                   cent_start-start, 
                                                   ideogram_height, 
                                           facecolor=facecolor,
                                           edgecolor='none',
                                           zorder=0,
                                           linewidth=linewidth,
                                           **kwargs
                                          ))
                g = ax.add_patch(patches.Rectangle((start, ideogram_base), 
                                                   cent_start-start, 
                                                   ideogram_height, 
                                           facecolor='none',
                                           edgecolor=edgecolor,
                                           zorder=0.5,
                                           linewidth=linewidth,
                                           **kwargs
                                          )) 
                g = ax.add_patch(patches.Rectangle((cent_end, ideogram_base), 
                                                   end-cent_end, 
                                                   ideogram_height, 
                                           facecolor=facecolor,
                                           edgecolor='none',
                                           zorder=0,
                                           linewidth=linewidth,
                                           **kwargs
                                          ))
                g = ax.add_patch(patches.Rectangle((cent_end, ideogram_base), 
                                                   end-cent_end, 
                                                   ideogram_height, 
                                           facecolor='none',
                                           edgecolor=edgecolor,
                                           zorder=0.5,
                                           linewidth=linewidth,
                                           **kwargs
                                          ))                   


                for ax in self.zoom_axes:

                    ideogram_base = self.map_y(base, ax)
                    ideogram_height = self.map_y(height, ax)
                
                    # draw centromere
                    ymin, ymax = ax.get_ylim()
                    if chrom in self.centromeres:
                        cent_start, cent_end = self.centromeres[chrom]
                    else:
                        cent_start, cent_end = 0, 0 

                    xy = [[cent_start, ideogram_base], 
                        [cent_start, ideogram_base+ideogram_height], 
                        [cent_end, ideogram_base], 
                        [cent_end, ideogram_base+ideogram_height]]

                    g = ax.add_patch(patches.Polygon(xy, closed=True, zorder=-1, 
                                                     fill=True, color='#777777'))
                    
                    # draw chrom
                    g = ax.add_patch(patches.Rectangle((start, ideogram_base),
                                                       cent_start-start,
                                                       ideogram_height, 
                                               facecolor=facecolor,
                                               edgecolor='none',
                                               zorder=-1,
                                               linewidth=linewidth,
                                               **kwargs
                                              ))
                    g = ax.add_patch(patches.Rectangle((start, ideogram_base), 
                                                       cent_start-start, 
                                                       ideogram_height, 
                                               facecolor='none',
                                               edgecolor=edgecolor,
                                               zorder=0.5,
                                               linewidth=linewidth,
                                               **kwargs
                                              )) 
                    g = ax.add_patch(patches.Rectangle((cent_end, ideogram_base), 
                                                       end-cent_end, 
                                                       ideogram_height, 
                                               facecolor=facecolor,
                                               edgecolor='none',
                                               zorder=0,
                                               linewidth=linewidth,
                                               **kwargs
                                              ))
                    g = ax.add_patch(patches.Rectangle((cent_end, ideogram_base), 
                                                       end-cent_end, 
                                                       ideogram_height, 
                                               # fill=False,
                                               facecolor='none',
                                               edgecolor=edgecolor,
                                               zorder=0.5,
                                               linewidth=linewidth,
                                               **kwargs
                                              ))               
        
    
    def _is_polygons_intersecting(self, a:Polygon, b:Polygon) -> bool:
        """
        Tests if two polygons intersect.

        Parameters
        ----------
        a : 
            A polygon
        b : 
            Another polygon

        Returns
        -------
        :
            True if polygons intersect, False otherwise
        """
        for x in range(2):
            polygon = a if x == 0 else b
    
            for i1 in range(len(polygon.get_points())):
                i2 = (i1 + 1) % len(polygon.get_points())
                p1 = polygon.get_points()[i1]
                p2 = polygon.get_points()[i2]
    
                normal = Point(p2.y - p1.y, p1.x - p2.x)
    
                min_a = float('inf')
                max_a = float('-inf')
    
                for p in a.get_points():
                    projected = normal.x * p.x + normal.y * p.y
                    min_a = min(min_a, projected)
                    max_a = max(max_a, projected)
    
                min_b = float('inf')
                max_b = float('-inf')
    
                for p in b.get_points():
                    projected = normal.x * p.x + normal.y * p.y
                    min_b = min(min_b, projected)
                    max_b = max(max_b, projected)
    
                if max_a < min_b or max_b < min_a:
                    return False
    
        return True

    
    def _scaled_y_lim(self, ax:matplotlib.axes.Axes, zoom=False) -> tuple:
        """
        Returns y-axis limits in plotting coordinates.

        Parameters
        ----------
        ax : 
            Matplotlib axes object

        Returns
        -------
        :
            Y-axis limits in plotting coordinates as tuple
        """
        # xlim = ax.get_xlim() - self.end_padding

        # bbox = ax.get_window_extent().transformed(self.fig.dpi_scale_trans.inverted())
        # aspect = bbox.height / bbox.width
        
        # return 0, -sub(*xlim) * aspect 

        if zoom:
            x_span = -sub(*ax.get_xlim()) #* (-sub(*ax.get_xlim()) - 2*self.end_padding) / -sub(*ax.get_xlim())  #+ int(zoom) * 2 * self.end_padding
        else:
            x_span = -sub(*ax.get_xlim())

        bbox = ax.get_window_extent().transformed(self.fig.dpi_scale_trans.inverted())
        aspect = bbox.height / bbox.width
        
        return 0, x_span * aspect 

        # return xlim[0] * self.aspect, xlim[1] * self.aspect
    

    def map_y(self, y:float, ax:matplotlib.axes.Axes, bottom:float=0, 
              top:float=1) -> float:
        """
        Maps y-axis values from user-specified ylim to actual plotting coordinates.

        Parameters
        ----------
        y : 
            Y-axis value
        ax : 
            Matplotlib axes object

        Returns
        -------
        :
            y-axis value in plotting coordinates
        """
        miny, maxy = ax.get_ylim()

        return (y * (top - bottom) * (maxy - miny) 
                / (self.ylim[1] - self.ylim[0]) + bottom * (maxy - miny))
    # zero = -miny
    # if y >= 0:
    #     return y * (top - max(bottom, zero)) * (maxy - zero) \
    #         / (self.ylim[1] - self.ylim[0]) + max(bottom, zero) * (maxy - zero)
    # else:
    #     return y * (min(top, zero) - bottom) * (zero - miny) \
    #         / (self.ylim[1] - self.ylim[0]) + min(top, zero) * (zero - miny)

        
    def draw_text(self, coord:float, x_pos:float, y_pos:float, text:str, 
                  textcolor:str, fontsize:float, linecolor:str, 
                  ax:matplotlib.axes.Axes=None, y_line_bottom:float=0, 
                  highlight:dict=None, **kwargs:dict) -> None:
        """
        Draws text with a line pointing to a position on the y-axis.

        Parameters
        ----------
        coord : 
            Coordonate of annotation
        x_pos : 
            X-axis position of text
        y_pos : 
            Y-axis position of text
        text : 
            Text to display
        textcolor : 
            Text color, by default 'black'
        linecolor : 
            Line color, by default 'lightgray'
        ax : 
            Matplotlib axes, by default None
        y_line_bottom : 
            y coordinate for bottom of line, by default 0
        highlight : 
            Dictionary for styling text labels, by default None
        """

        y_unit = -sub(*self._scaled_y_lim(ax)) / -sub(*self.ylim)

        if 'color' in kwargs:
            textcolor = kwargs['color']
            linecolor = kwargs['color']
            del kwargs['color']

        kwargs.setdefault('zorder', 10)
        kwargs.setdefault('linewidth', 0.5)

        if highlight:
            text_props = highlight
        else:
            text_props = dict(color=textcolor,
                    # 'normal' | 'small-caps':
                    #variant = 'small-caps', 
                    # 'normal' | 'bold' | 'heavy' | 'light' | 'ultrabold' | 'ultralight':
                    fontweight='normal', 
                    # 'normal' | 'italic' | 'oblique'
                    fontstyle = 'normal', 
                    bbox=dict(boxstyle='square,pad=0', 
                              linewidth=0.2,
                              fc='none', 
                              alpha=1,
                              ec='none'),)

        t = ax.text(x_pos, y_pos, text, fontsize=fontsize,                     
                    rotation=45, zorder=10, 
                    horizontalalignment='left',
                    verticalalignment='bottom',
                    **text_props)


        ax.plot((coord, coord, x_pos+y_unit/5),
                (y_line_bottom, y_pos, y_pos+y_unit/5),
                solid_capstyle='butt', 
                solid_joinstyle='miter',
                color=linecolor, 
                **kwargs)
    
    
    def get_polygon(self, text:str, x_pos:int, y_pos:float,
                    ax:matplotlib.axes.Axes, pad=0) -> matplotlib.patches.Polygon:
        """
        Computes polygon for rotated text label.

        Parameters
        ----------
        text : 
            Text to display
        x_pos : 
            X-axis position
        y_pos : 
            Y-axis position
        ax : 
            Matplotlib axes
        pad : 
            Text padding, by default 0

        Returns
        -------
        :
            Maplotlib polygon object
        """

        y_unit = -sub(*self._scaled_y_lim(ax)) / -sub(*self.ylim)
        # y_unit = -sub(*self.scaled_y_lim(ax)) / -sub(*ax.get_ylim())
        # y_pos = y_pos * y_unit

        y_pos = self.map_y(y_pos, ax)
        
        t = ax.text(x_pos, y_pos, text, fontsize=self.zoom_font_size,
                    horizontalalignment='left',
                    verticalalignment='bottom', 
                    rotation=0, zorder=3, 
                    bbox=dict(boxstyle='square', 
                              fc='none', ec='red', pad=pad, alpha=0.4))
        
        transf = ax.transData.inverted()
        bb = t.get_window_extent(renderer = self.fig.canvas.get_renderer())
        bbt = bb.transformed(transf)
        coords = bbt.get_points()
        
        Artist.remove(t)
        
        (x0, y0), (x1, y1) = coords
        pad = (y1 - y0)/5
        coords = np.array([(x0, y0), (x0, y1), (x1, y1), (x1, y0)])
            
        x, y = zip(*coords)
        df = pd.DataFrame(dict(x=x, y=y))
        df['y'] -= y_pos
        df['x'] -= x_pos
        dfx = df['x']*math.cos(math.radians(45))-df['y']*math.sin(math.radians(45))
        df['y'] = (df['x'] * math.sin(math.radians(45)) + df['y'] 
                   * math.cos(math.radians(45)))
        df['x'] = dfx
        df['y'] += y_pos
        df['x'] += x_pos

        _x = np.sort(df.x.to_numpy())
        df['x'] += _x[1] - _x[0]
        _y = np.sort(df.y.to_numpy())
        df['y'] -= (_y[1] - _y[0])/2

        coords = np.array(list(zip(df.x, df.y)))
        
        x_pos, y_pos = df['x'][0] - (df['x'][0] - df['x'][1]), df['y'][0]
    
        df['x'] = df.x + np.array([-pad, -pad, pad, pad])
        df['y'] = df.y + np.array([-pad, pad, pad, -pad])
        
        return x_pos, y_pos, Polygon([Point(x, y) for x, y in zip(df.x, df.y)])
        

    # def add_labels(self, data, labels='name', chrom='chrom', x='pos'):
    def add_labels(self, annot:MutableSequence, 
                   base:float=None, min_height:float=None, 
                   zoom_base:float=None, zoom_min_height:float=None, 
                   bold:MutableSequence=[], 
                   italic:MutableSequence=[], colored:MutableSequence=[], 
                   framed:MutableSequence=[], filled:MutableSequence=[], 
                   highlight_color='deeppink',
                   pad:float=0,  **kwargs:dict) -> None:
        """
        Add text labels to the chromosome ideograms. 

        Parameters
        ----------
        annot : 
            List of gene names or tuples of HGCN gene symbols and tuples with
            chromosome name, gene position, text label, and optionally text color,
            text size, and line color: `(<chrom>, <position>, <text>, [<textcolor>, 
            [ [textsize] [<linecolor>] ] ])`. Text size values are scaled to make 
            the largest one 1 and font size is then computed as self.font_size * 
            textsize. That way font sizes can only be smaller than the default.
        base : 
            Y coordinate for lower end of vertical line, by default None. 
            If None, the upper edge of ideogram is used.
        min_height : 
            Minimum length of vertical line in y coordinates, by default None. 
            If None, half the ideogram height is used.
        bold : 
            List of genes to highlight with bold text, by default []
        italic : 
            List of genes to highlight with italic text, by default []
        colored : 
            List of genes to highlight with color, by default []
        framed : 
            List of genes to highlight with framed label, by default []
        filled : 
            List of genes to highlight with filled label, by default []
        highlight_color : 
            Color used for highlighted genes, by default 'deeppink'
        pad : 
            Text padding, by default 0
        zoom_base : 
            Y coordinate for lower end of vertical line on zoom axes, by default None. 
            If None, the default value of font_size.
        zoom_min_height : 
            Minimum length of vertical line in y coordinate on zoom axes, by default None. 
            If None, the default is the value of min_height..
        """
        
        if base is None:
            base = self.ideogram_base + self.ideogram_height 

        if min_height is None:
            min_height = self.ideogram_height * 0.5

        if zoom_base is None:
            zoom_base = base

        if zoom_min_height is None:
            zoom_min_height = min_height

        if type(annot[0]) is str:
            _annot = []
            # annot is a list of gene names, not a list of tuples
            for gene_name, (chrom, start, end, strand) in gene_coord(
                annot, assembly=self.assembly).items():
                _annot.append((chrom, (start + end)/2, gene_name))
            annot = _annot

        _annot = []
        for a in annot:
            if len(a) == 3:
                a = a + (None, 1.0, None)
            elif len(a) == 4:
                a = a + (1.0, None)
            elif len(a) == 5:
                a = a + (None,)
            elif len(a) == 6:
                pass
            else:
                raise ValueError('Invalid number of elements in annotation tuple')
            _annot.append(a)
        annot = _annot

        max_textsize = max([t[4] for t in annot])
        annot = [(a[0], a[1], a[2], a[3], a[4]/max_textsize, a[5]) for a in annot]

        y0 = base
        y1 = base + min_height

        zoom_y0 = zoom_base
        zoom_y1 = zoom_base + zoom_min_height
        
        chrom_annot = defaultdict(list)
        for a in annot:
            chrom_annot[a[0]].append(a[1:])
            
        for chrom, annot in chrom_annot.items():
            if chrom not in self.chr_axes:
                continue
            ax = self.chr_axes[chrom]

            foreground_color = black_white(ax)            

            hsv = matplotlib.colors.rgb_to_hsv(matplotlib.colors.to_rgb(highlight_color))
            hsv[1] *= 0.15
            highlight_fill = matplotlib.colors.hsv_to_rgb(hsv)

            highlight = defaultdict(dict)
            for gene in bold:
                highlight[gene].update(dict(weight='bold'))
            for gene in italic:
                highlight[gene].update(dict(style='italic'))
            for gene in colored:
                highlight[gene].update(dict(color=highlight_color))
            for gene in framed:
                if 'bbox' not in highlight[gene]:
                    highlight[gene]['bbox'] = {}            
                highlight[gene]['bbox'].update(dict(edgecolor=foreground_color, 
                                                    pad=max(1.5, pad), 
                                                    linewidth=0.5))
            for gene in filled:
                if 'bbox' not in highlight[gene]:
                    highlight[gene]['bbox'] = {}
                highlight[gene]['bbox'].update(dict(facecolor=highlight_fill,
                                                    pad=max(1.5, pad)))

            for gene in highlight:
                if 'bbox' in highlight[gene]:
                    if 'edgecolor' not in highlight[gene]['bbox']:
                        highlight[gene]['bbox']['edgecolor'] = 'none'
                    if 'facecolor' not in highlight[gene]['bbox']:
                        highlight[gene]['bbox']['facecolor'] = 'none'


            annot = sorted(annot, reverse=True)

            y_unit = -sub(*self._scaled_y_lim(ax)) / -sub(*self.ylim)
            nudge = 0.01 * y_unit
        
            polybuff = []
            for pos, name, textcolor, textsize, linecolor in annot:

                if textcolor is None:
                    textcolor = foreground_color
                if linecolor is None:
                    linecolor = foreground_color

                if type(highlight) is list or type(highlight) is set:
                    hl = name in highlight
                elif type(highlight) is dict or type(highlight) is defaultdict:
                    hl = highlight[name]
                else:
                    hl = None
                
                x, y, poly = self.get_polygon(name, pos, y1, ax, pad=pad)
                
                while any(self._is_polygons_intersecting(poly, p) 
                          for p in polybuff):
                    y += nudge
                    poly.nudge_y(nudge)

                self.draw_text(pos, x, y, name, textcolor=textcolor, 
                               fontsize=textsize*self.font_size, linecolor=linecolor,
                            #    ax=ax, y_line_bottom=y0*y_unit,                                
                               ax=ax, y_line_bottom=self.map_y(y0, ax),                                
                               highlight=hl, **kwargs)

                polybuff.append(poly)
                if len(polybuff) > 40:
                    del polybuff[0]        
            z = 100
            for i, t in enumerate(reversed(ax.texts)):
                t.set_zorder(z+i)

            for zoom, ax in zip(self.zooms, self.zoom_axes):

                # y_unit = -sub(*self._scaled_y_lim(ax, zoom=True)) / -sub(*self.ylim)
                # nudge = 0.01 * y_unit
                self.map_y(0.01, ax)
        
                zoom_polybuff = []
                for pos, name, *args in annot:

                    if type(highlight) is list or type(highlight) is set:
                        hl = name in highlight
                    elif type(highlight) is dict or type(highlight) is defaultdict:
                        hl = highlight[name]
                    else:
                        hl = None
                        
                    if pos >= zoom[0] and pos < zoom[1]:
                        x, y, poly = self.get_polygon(name, pos, zoom_y1, ax)
                        while any(self._is_polygons_intersecting(poly, p) 
                                  for p in zoom_polybuff):
                            y += nudge
                            poly.nudge_y(nudge)

                        self.draw_text(pos, x, y, name, textcolor=textcolor, 
                               fontsize=textsize*self.zoom_font_size, linecolor=linecolor,
                            #    ax=ax, y_line_bottom=y0*y_unit,                                
                               ax=ax, y_line_bottom=self.map_y(zoom_y0, ax),                                 
                               highlight=hl, **kwargs)
        
                        zoom_polybuff.append(poly)
                        if len(zoom_polybuff) > 20:
                            del zoom_polybuff[0]        
                z = 100
                for i, t in enumerate(reversed(ax.texts)):
                    t.set_zorder(z+i)
                    


    def add_segments(self, annot:MutableSequence, base:float=None, 
                     height:float=None,label:str=None, min_visible_width:int=200000, 
                     **kwargs:dict) -> None:
        """
        Add segments (rectangles) to the chromosome ideograms.

        Parameters
        ----------
        annot : 
            List of tuples with chromosome name, start and end positions of each
            segment: `(<chrom>, <start>, <end>, [<alpha>])`
        base : 
            Y coordinate for lower edge of rectangles, by default None. 
            If None, the lower edge of ideogram is used.
        height : 
            Height of rectangles in y coordinates, by default None. 
            If None, the ideogram height is used.
        label : 
            Label for plot legend, by default None
        min_visible_width : 
            Minimum with of rendered rectangles ensuring that very short segments
            remain visible, by default 200000
        **kwargs :
            Additional keyword arguments for matplotlib.patches. Rectangle
            (defaults to {facecolor='black', edgecolor='none'})
        """

        kwargs.setdefault('facecolor', 'black')
        kwargs.setdefault('edgecolor', 'none')
        
        if label is not None:
            self.legend_handles.append(
                patches.Patch(facecolor=kwargs['facecolor'], 
                              edgecolor=kwargs['edgecolor'],label=label)
            )
        
        if base is None:
            base = self.ideogram_base
        if height is None:
            height = self.ideogram_height            

        chrom_annot = defaultdict(list)
        for a in annot:
            chrom_annot[a[0]].append(a[1:])
            
        for chrom, annot in chrom_annot.items():
            ax = self.chr_axes[chrom]
            annot = sorted(annot, reverse=True)
            y_unit = -sub(*self._scaled_y_lim(ax)) / -sub(*self.ylim)
            for start, end, *extra in annot:
                _kwargs = kwargs.copy()
                if extra and 'alpha' not in kwargs:
                    _kwargs['alpha'] = extra[0]
                else:
                    _kwargs['alpha'] = 1
                # scaled_base = base * y_unit
                # scaled_height = height * y_unit                
                scaled_base = self.map_y(base, ax)
                scaled_height = self.map_y(height, ax)
                width = end - start
                if width < min_visible_width:
                    start -= min_visible_width/2
                    width += min_visible_width

                rect = patches.Rectangle((start, scaled_base), width, 
                                         scaled_height, linewidth=1, zorder=0,
                                         **_kwargs)
                ax.add_patch(rect)    
                
                for zoom_ax in self.zoom_axes:
                    zoom_y_unit = -sub(*self._scaled_y_lim(zoom_ax, zoom=True)) / -sub(*self.ylim)
                    # zoom_scaled_base = base * zoom_y_unit
                    # zoom_scaled_height = height * zoom_y_unit                      
                    zoom_scaled_base = self.map_y(base, zoom_ax)
                    zoom_scaled_height = self.map_y(height, zoom_ax)               
                    # print(start, zoom_scaled_base, width,
                    #                          zoom_scaled_height)
                    rect = patches.Rectangle((start, zoom_scaled_base), width,
                                             zoom_scaled_height, linewidth=1,
                                             zorder=0, **_kwargs)
                    zoom_ax.add_patch(rect)    


    def bezier_lines(self, ax, pairs, base, height, adjust_heights,
                     color, color_cycle, linewidth, zorder, **kwargs):
    
        def _get_color(n, lightness=0.4):
            color_cycle = cycle([matplotlib.colors.to_hex(c) for c in sns.husl_palette(n, l=lightness)])
            for color in color_cycle:
                yield color
        
        if color_cycle is None:
            husl_colors = _get_color(len(pairs), lightness=0.7)
        else:
            husl_colors = _get_color(color_cycle, lightness=0.7)
        
        bezier_path = np.linspace(0, 1, 100)

        for x1, x2 in pairs:

            y1, y2 = base, base
            if adjust_heights:
                h = self.map_y(height, ax) * np.log(abs(x1-x2)) / 10
            else:
                h = self.map_y(height, ax) * 2
    
            # Control point: midpoint lifted vertically
            xm, ym = (x1 + x2) / 2, (y1 + y2) / 2 + h
        
            # Compute Bézier curve points
            x = (1 - bezier_path) ** 2 * x1 + 2 * (1 - bezier_path) * bezier_path * xm + bezier_path ** 2 * x2
            y = (1 - bezier_path) ** 2 * y1 + 2 * (1 - bezier_path) * bezier_path * ym + bezier_path ** 2 * y2
        
            y += self.map_y(base, ax) 

            if color is None:
                _color = next(husl_colors)
            else:
                _color = color
            ax.plot(x, y, '-', linewidth=linewidth, color=_color, 
                    zorder=zorder, **kwargs)


    def add_connections(self, annot:MutableSequence, base:float=None, 
                    height:float=None, color:str=None, color_cycle=None, linewidth:float=1,
                    zorder:float=100, adjust_heights=True, **kwargs:dict) -> None:

        if base is None:
            base = self.ideogram_base + self.ideogram_height
        if height is None:
            height = self.ideogram_height * 2

        chrom_annot = defaultdict(list)
        for a in annot:
            chrom_annot[a[0]].append(a[1:])

        for chrom, annot in chrom_annot.items():
            if chrom not in self.chr_axes:
                continue
            ax = self.chr_axes[chrom]
            annot = sorted(annot, reverse=True)
        
            self.bezier_lines(ax, annot, base=base, height=height, 
                              adjust_heights=adjust_heights, color=color,
                              color_cycle=color_cycle,
                              linewidth=linewidth, zorder=zorder, **kwargs)
            
            for ax in self.zoom_axes:
                self.bezier_lines(ax, annot, base=base, height=height, 
                              adjust_heights=adjust_heights, color=color,
                              color_cycle=color_cycle,
                              linewidth=linewidth, zorder=zorder, **kwargs)
    

    def add_vlines(self, step:int=1000000, color:str='black', linewidth:float=0.1,
                   zorder:float=100, **kwargs:dict) -> None:
        """
        Adds vertical lines to the chromosome ideograms.

        Parameters
        ----------
        step : 
            Number of bases between lines, by default 1000000
        color : 
            Color of lines, by default 'black'
        linewidth : 
            Width of lines, by default 0.1
        zorder : 
            zorder of lines, by default 100
        """
        for i, ax in enumerate(self.ax_list):

            s, e = self.centromeres[self.chr_names[i]]

            ymin = self.map_y(self.ideogram_base, ax)        
            ymax = self.map_y(self.ideogram_base+self.ideogram_height, ax)
            ax.vlines(
                [x for x in range(0, self.chr_sizes[i], step) if x < s or x > e],
                ymin=ymin, ymax=ymax, linewidth=linewidth, color=color,
                zorder=zorder, **kwargs)
            for ax in self.zoom_axes:
                ymin = self.map_y(self.ideogram_base, ax)        
                ymax = self.map_y(self.ideogram_base+self.ideogram_height, ax)
                ax.vlines(
                    [x for x in range(0, self.chr_sizes[i], step) if x < s or x > e], 
                    ymin=ymin, ymax=ymax, linewidth=linewidth, color=color,
                    zorder=zorder, **kwargs)
                
    
    # TODO rename yaxis to data_scaling or something or use offset and scale 
    # or offset and height
    def map_method(self, method:Callable, data:pd.DataFrame=None, ch:str='chrom',
                   yaxis:tuple=(0.5, 3.5), **kwargs:dict) -> None:
        """
        Map a matplotib axes method like plot and scatter to each ideogram 
        panel (axis).
        
        Parameters
        ----------
        method : 
            Method of matplotib.axes.Axes to apply to each ideogram panel 
            (E.g. Axes.scatter).
        data : 
            Data frame with with data for x and y coordinates by chromsome
        ch : 
            Name of data frame column holding chromosome names, by default 
            'chrom'
        x : str
            Name of data frame column holding x coordinates, by default 'x'
        y : str
            Name of data frame column holding y coordinates, by default 'y'
        yaxis : 
            Y interval of ideogram panel axis ideogram to map data to, 
            by default (0.5, 3.5)
        **kwargs :
            Additional keyword arguments are passed to the plotting function 
            as keyword arguments.
        """
        def method_not_found(): # just in case we dont have the function
            print('No Function '+method+' Found!')

        grouped = data.groupby(ch, observed=True)
        for chrom, group in grouped:
            if chrom not in self.chr_axes:
                continue
            ax = self.chr_axes[chrom]
            bottom, top = yaxis            
            scaled_y_lim = ax.get_ylim()
            if 'ylim' in kwargs:
                dy = -sub(*kwargs['ylim'])
            else:
                dy = -sub(*self.ylim)
            df = group.reset_index() # create independent dataframe
            df['y'] -= df.y.min()
            df['y'] /= df.y.max()
            df['y'] = (df.y * ((top-bottom) * -sub(*scaled_y_lim)) 
                       / -sub(*self.ylim) + bottom 
                       /  -sub(*self.ylim) * -sub(*scaled_y_lim))
            x = df.x
            y = df.y
            if 'x' in kwargs: del kwargs['x']
            if 'y' in kwargs: del kwargs['y'] 

            if 'y2' in kwargs:
                if 'y2' not in df.columns:
                    df['y2'] = kwargs['y2']
                df['y2'] -= df.y2.min()
                df['y2'] /= df.y2.max()
                df['y2'] = (df.y2 * ((top-bottom) * -sub(*scaled_y_lim)) 
                        / -sub(*self.ylim) + bottom 
                        /  -sub(*self.ylim) * -sub(*scaled_y_lim))
                y2 = df.y2

            method_name = method.__name__
            method = getattr(ax, method_name, method_not_found) 
            g = method(x, y, **kwargs)
            
            for ax in self.zoom_axes:
                scaled_y_lim = ax.get_ylim()
                if 'ylim' in kwargs:
                    dy = -sub(*kwargs['ylim'])
                else:
                    dy = -sub(*self.ylim)
                df = group.reset_index() # create independent dataframe
                df['y'] -= df.y.min()
                df['y'] /= df.y.max()
                df['y'] = (df.y * ((top-bottom) * -sub(*scaled_y_lim)) 
                           / -sub(*self.ylim) + bottom 
                           /  -sub(*self.ylim) * -sub(*scaled_y_lim))
                x = df.x
                y = df.y
                if 'x' in kwargs: del kwargs['x']
                if 'y' in kwargs: del kwargs['y']   
                
                method = getattr(ax, method_name, method_not_found) 
                g = method(x, y, **kwargs)
                
                try:
                    ax.get_legend().remove()
                except:
                    pass
                ax.set_xlabel('')
                ax.set_ylabel('')  
            plt.xlabel('')
            plt.ylabel('')

    def map_fun(self, fun:Callable, data:pd.DataFrame=None, ch='chrom', 
                yaxis:tuple=(0.5, 3.5), **kwargs:dict) -> None:
        """
        Map a plotting function like seaborn.scatterplot to each ideogram panel (axis).
        
        Parameters
        ----------
        fun : 
            Function to apply to each ideogram panel (E.g. seaborn.scatterplot).
        data : 
            Data frame with with data for x and y coordinates by chromsome.
        ch : 
            Name of data frame column holding chromosome names, by default 'chrom'
        x : str
            Name of data frame column holding x coordinates, by default 'x'
        y : str
            Name of data frame column holding y coordinates, by default 'y'
        yaxis : 
            Y interval of ideogram panel axis ideogram to map data to, 
            by default (0.5, 3.5)
        **kwargs :
            Additional keyword arguments are passed to the plotting function as
            keyword arguments.            
        """            
        grouped = data.groupby(ch, observed=True)
        for chrom, group in grouped:
            if chrom not in self.chr_axes:
                continue
            ax = self.chr_axes[chrom]
            scaled_y_lim = ax.get_ylim()
            if 'ylim' in kwargs:
                dy = -sub(*kwargs['ylim'])
            else:
                dy = -sub(*self.ylim)
            df = group.reset_index() #.copy()
            if yaxis is None:
                df['y'] = df.y * -sub(*scaled_y_lim) / -sub(*self.ylim)
            else:
                bottom, top = yaxis
                df['y'] -= df.y.min()
                df['y'] /= df.y.max()
                df['y'] = (df.y * ((top-bottom) * -sub(*scaled_y_lim)) 
                           / -sub(*self.ylim) + bottom 
                           /  -sub(*self.ylim) * -sub(*scaled_y_lim))
            g = fun(df, ax=ax, **kwargs)
            
            for ax in self.zoom_axes:
                scaled_y_lim = ax.get_ylim()
                if 'ylim' in kwargs:
                    dy = -sub(*kwargs['ylim'])
                else:
                    dy = -sub(*self.ylim)
                df = group.reset_index() #.copy()
                if yaxis is None:
                    df['y'] = df.y * -sub(*scaled_y_lim) / -sub(*self.ylim)
                else:
                    bottom, top = yaxis
                    df['y'] -= df.y.min()
                    df['y'] /= df.y.max()
                    df['y'] = (df.y * ((top-bottom) * -sub(*scaled_y_lim)) 
                               / -sub(*self.ylim) + bottom 
                               / -sub(*self.ylim) * -sub(*scaled_y_lim))
                g = fun(df, ax=ax, **kwargs)
                try:
                    ax.get_legend().remove()
                except:
                    pass
                ax.set_xlabel('')
                ax.set_ylabel('')                               
            plt.xlabel('')
            plt.ylabel('')

    def add_legend(self, outside=False, **kwargs:dict) -> None:
        """
        Adds a legend to the chromosome ideograms.

        Parameters
        ----------
        outside : 
            Place the legend outside the plot, by default False
        **kwargs :
            Keyword arguments are passed to matplotlib's `legend`. 
            Defaults to {'loc': 'center left', 'bbox_to_anchor': (1.02, 0.5),
            'frameon': False} if `outside` is True, otherwise defaults to {'frameon': False}.
        """
        if outside:
            kwargs.setdefault('loc', 'center left')
            kwargs.setdefault('bbox_to_anchor', (1.02, 0.5))
        kwargs.setdefault('frameon', False)

        for i, ax in enumerate(self.ax_list):
            if i == 0:                
                handles, labels = ax.get_legend_handles_labels()
                if self.legend_handles:
                    handles = self.legend_handles + handles
                ax.legend(handles=handles, **kwargs)
            else:
                ax.legend_.remove()



    def add_horizon(self, data:pd.DataFrame=None, ch:str='chrom', y:str='y', x:str='x', 
                    cut:float=None, # float, takes precedence over quantile_span
                    quantile_span:float=None, beginzero:bool=True, 
                    base:float=None, height:float=None,
                    colors:List[str] = [
                        '#CCE2DF', '#59A9A8', '#374E9B', 'midnightblue',
                        '#F2DE9A', '#DA8630', '#972428', 'darkred',
                        '#D3D3D3'],
                    **kwargs:dict) -> None:
        """
        Adds a horizon plot to the chromosome ideograms.

        Parameters
        ----------
        data : 
            Pandas DataFrame with sorted x and y data grouped by chromosome
        ch : 
            Name of data frame column holding chromosome names, by default 'chrom'
        x : 
            Name of data frame column holding x coordinates, by default 'x'
        y : 
            Name of data frame column holding y coordinates, by default 'y'
        cut : 
            Lower and upper y values for folding the horizon plot. Default for each
            chromosome is a third of the y range on each side of zero.
        quantile_span : 
            Lower and upper quantiles of y values for each chromosome to include 
            in the horizon plot. `quantile_span=(0, 1)` produces the same result 
            as `cut=None`. Outlier values below and above this range is assigned
            separate darker colors. 
        beginzero : 
            Pad to make the x axis begin at zero, by default True
        base : 
            Y coordinate for lower edge of horizon plot, by default None.
            If None, the lower edge of ideogram is used.
        height : 
            Height of horizon plot in y coordinates, by default None. 
            If None, the ideogram height is used.
        colors : 
            _description_, by default [ '#CCE2DF', '#59A9A8', '#374E9B',
            'midnightblue', '#F2DE9A', '#DA8630', '#972428', 'darkred', '#D3D3D3']
        """
        if base is None:
            if self.ideogram_base is not None:
                base = self.ideogram_base
            else:
                base = 0
        if height is None:
            if self.ideogram_height is not None:
                height = self.ideogram_height
            else:
                height = 1

        grouped = data.groupby(ch, observed=True)
        for chrom, group in grouped:
            if chrom not in self.chr_axes:
                continue
            ax = self.chr_axes[chrom]
            
            df = group.reset_index() # make copy
            horizon(df, y=y, ax=ax,
                    cut=cut, 
                    quantile_span = quantile_span,
                    x=x,
                    beginzero=beginzero, 
                    offset=self.map_y(base, ax),
                    height=self.map_y(height, ax),
                    colors = colors,
                    **kwargs)
                             
            for ax in self.zoom_axes:
                horizon(df, y=y, ax=ax,
                    cut=cut, 
                    quantile_span = quantile_span,
                    x=x,
                    beginzero=beginzero, 
                    offset=self.map_y(base, ax),
                    height=self.map_y(height, ax),
                    colors = colors,
                    **kwargs)
    

    def annotate(self, xy:tuple=None, testxy:tuple=None, text:str=None, 
                 chrom:str=None, ax:matplotlib.axes.Axes=None, wrap:int=30, 
                 **kwargs:dict) -> None:
        """
        Draws an annotation with an arrow to a point from a text box.

        Parameters
        ----------
        xy : 
            Coordinate to point to as a tuple of chromosome position and 
            y plotting coordinate
        testxy : 
            Coordinate to text box as a tuple of chromosome position and 
            y plotting coordinate
        text : 
            Text for text box
        chrom : 
            Chromosome to annotate, by default the first or only chromosome 
            in the plot
        ax : 
            Axis with the point the arrow points to, by default the axis for 
            the chromosome ideogram.
        wrap : 
            Line wrap for text box, by default 30
        kwargs : 
            Additional keyword arguments passed to matplotlib's annotate
        """
        kwargs.setdefault('fontsize', 7)
        kwargs.setdefault('ha', 'left')
        kwargs.setdefault('va', 'bottom')
        kwargs.setdefault(
            'bbox', dict(facecolor='white', edgecolor='red', linewidth=1))
        kwargs.setdefault('arrowprops', dict(facecolor='red', # color of arrow
                                            shrink=0.01, # distance from point
                                            width=1.5, # width of arrow
                                            headwidth=5, # width of arrow head
                                            headlength=7, # length of arrow head
                                            connectionstyle='angle3,angleA=0,angleB=-90'))
        if chrom is None:
            chrom = self.chr_names[0]
        _ax = self.chr_axes[chrom]
        if ax is None:
            ax = _ax
            xy = (xy[0], self.map_y(xy[1], _ax))

        dpi = self.fig.dpi
        bbox = _ax.get_window_extent()
        width_inches = bbox.width / dpi  # in inches
        point_in_bases = _ax.get_xlim()[1]/(width_inches*72)

        ax.annotate(textwrap.fill(text, wrap), 
                    xy=xy, 
                    xycoords="data",
                    xytext=(testxy[0]/point_in_bases, 
                            self.map_y(testxy[1], _ax)/point_in_bases), 
                    textcoords=OffsetFrom(_ax.bbox, (0, 0), "points"),
                    **kwargs) 




import sys
from functools import partial
import itertools
import random
from collections import defaultdict
import numpy as np
from matplotlib.pyplot import subplots
from matplotlib.patches import Ellipse, Polygon, Patch
from matplotlib.colors import to_rgba
from matplotlib.cm import ScalarMappable

SHAPE_COORDS = {
    2: [(.375, .500), (.625, .500)],
    # 3: [(.333, .633), (.666, .633), (.500, .310)],
    3: [(.333, .633), (.666, .633), (.500, .310+0.07)],
    4: [(.350, .400), (.450, .500), (.544, .500), (.644, .400)],
    5: [(.428, .449), (.469, .543), (.558, .523), (.578, .432), (.489, .383)],
    6: [
        (.637, .921, .649, .274, .188, .667),
        (.981, .769, .335, .191, .393, .671),
        (.941, .397, .292, .475, .456, .747),
        (.662, .119, .316, .548, .662, .700),
        (.309, .081, .374, .718, .681, .488),
        (.016, .626, .726, .687, .522, .327)
    ]
}

SHAPE_DIMS = {
    2: [(.50, .50), (.50, .50)],
    3: [(.50, .50), (.50, .50), (.50, .50)],
    4: [(.72, .45), (.72, .45), (.72, .45), (.72, .45)],
    5: [(.87, .50), (.87, .50), (.87, .50), (.87, .50), (.87, .50)],
    6: [(None,)]*6
}

SHAPE_ANGLES = {
    2: [0, 0],
    3: [0, 0, 0],
    4: [140, 140, 40, 40],
    5: [155, 82, 10, 118, 46], 6: [None]*6
}

PETAL_LABEL_COORDS = {
    2: {
        "01": (.74, .50), "10": (.26, .50), "11": (.50, .50)
    },
    # 3: {
    #     "001": (.500, .270), "010": (.730, .650), "011": (.610, .460),
    #     "100": (.270, .650), "101": (.390, .460), "110": (.500, .650),
    #     "111": (.500, .508)
    # },
    3: {
        "001": (.500, .270 + .07), "010": (.730, .650), "011": (.610+0.02, .460 + .03),
        "100": (.270, .650), "101": (.390-0.02, .460 + .03), "110": (.500, .650+0.05),
        "111": (.500, .508 + .03)
    },    4: {
        "0001": (.85, .42), "0010": (.68, .72), "0011": (.77, .59),
        "0100": (.32, .72), "0101": (.71, .30), "0110": (.50, .66),
        "0111": (.65, .50), "1000": (.14, .42), "1001": (.50, .17),
        "1010": (.29, .30), "1011": (.39, .24), "1100": (.23, .59),
        "1101": (.61, .24), "1110": (.35, .50), "1111": (.50, .38)
    },
    5: {
        "00001": (.27, .11), "00010": (.72, .11), "00011": (.55, .13),
        "00100": (.91, .58), "00101": (.78, .64), "00110": (.84, .41),
        "00111": (.76, .55), "01000": (.51, .90), "01001": (.39, .15),
        "01010": (.42, .78), "01011": (.50, .15), "01100": (.67, .76),
        "01101": (.70, .71), "01110": (.51, .74), "01111": (.64, .67),
        "10000": (.10, .61), "10001": (.20, .31), "10010": (.76, .25),
        "10011": (.65, .23), "10100": (.18, .50), "10101": (.21, .37),
        "10110": (.81, .37), "10111": (.74, .40), "11000": (.27, .70),
        "11001": (.34, .25), "11010": (.33, .72), "11011": (.51, .22),
        "11100": (.25, .58), "11101": (.28, .39), "11110": (.36, .66),
        "11111": (.51, .47)
    },
    6: {
        "000001": (.212, .562), "000010": (.430, .249), "000011": (.356, .444),
        "000100": (.609, .255), "000101": (.323, .546), "000110": (.513, .316),
        "000111": (.523, .348), "001000": (.747, .458), "001001": (.325, .492),
        "001010": (.670, .481), "001011": (.359, .478), "001100": (.653, .444),
        "001101": (.344, .526), "001110": (.653, .466), "001111": (.363, .503),
        "010000": (.750, .616), "010001": (.682, .654), "010010": (.402, .310),
        "010011": (.392, .421), "010100": (.653, .691), "010101": (.651, .644),
        "010110": (.490, .340), "010111": (.468, .399), "011000": (.692, .545),
        "011001": (.666, .592), "011010": (.665, .496), "011011": (.374, .470),
        "011100": (.653, .537), "011101": (.652, .579), "011110": (.653, .488),
        "011111": (.389, .486), "100000": (.553, .806), "100001": (.313, .604),
        "100010": (.388, .694), "100011": (.375, .633), "100100": (.605, .359),
        "100101": (.334, .555), "100110": (.582, .397), "100111": (.542, .372),
        "101000": (.468, .708), "101001": (.355, .572), "101010": (.420, .679),
        "101011": (.375, .597), "101100": (.641, .436), "101101": (.348, .538),
        "101110": (.635, .453), "101111": (.370, .548), "110000": (.594, .689),
        "110001": (.579, .670), "110010": (.398, .670), "110011": (.395, .653),
        "110100": (.633, .682), "110101": (.616, .656), "110110": (.587, .427),
        "110111": (.526, .415), "111000": (.495, .677), "111001": (.505, .648),
        "111010": (.428, .663), "111011": (.430, .631), "111100": (.639, .524),
        "111101": (.591, .604), "111110": (.622, .477), "111111": (.501, .523)
    }
}

PSEUDOVENN_PETAL_COORDS = {
    6: {
        "100000": (.275, .875), "010000": (.725, .875), "001000": (.925, .500),
        "000100": (.725, .125), "000010": (.275, .125), "000001": (.075, .500),
        "110000": (.500, .850), "011000": (.800, .675), "001100": (.800, .325),
        "000110": (.500, .150), "000011": (.200, .325), "100001": (.200, .675),
        "110001": (.375, .700), "111000": (.625, .700), "011100": (.750, .500),
        "001110": (.625, .300), "000111": (.375, .300), "100011": (.250, .500),
        "111001": (.500, .650), "111100": (.635, .575), "011110": (.635, .415),
        "001111": (.500, .350), "100111": (.365, .415), "110011": (.365, .575),
        "111011": (.440, .600), "111101": (.560, .600), "111110": (.620, .500),
        "011111": (.560, .400), "101111": (.440, .400), "110111": (.380, .500),
        "111111": (.500, .500)
    }
}


def generate_colors(cmap="viridis", n_colors=6, alpha=.4):
    """Generate colors from matplotlib colormap; pass list to use exact colors"""
    if not isinstance(n_colors, int) or (n_colors < 2) or (n_colors > 6):
        raise ValueError("n_colors must be an integer between 2 and 6")
    if isinstance(cmap, list):
        colors = [to_rgba(color, alpha=alpha) for color in cmap]
    else:
        scalar_mappable = ScalarMappable(cmap=cmap)
        colors = scalar_mappable.to_rgba(range(n_colors), alpha=alpha).tolist()
    return colors[:n_colors]

def less_transparent_color(color, alpha_factor=2):
    """Bump up color's alpha"""
    new_alpha = (1 + to_rgba(color)[3]) / alpha_factor
    return to_rgba(color, alpha=new_alpha)

def draw_ellipse(ax, x, y, w, h, a, color):
    """Wrapper for drawing ellipse; called like `draw_ellipse(ax, *coords, *dims, angle, color)`"""
    ax.add_patch(
        Ellipse(
            xy=(x,y), width=w, height=h, angle=a,
            facecolor=color, edgecolor=less_transparent_color(color)
        )
    )

def draw_triangle(ax, x1, y1, x2, y2, x3, y3, _dim, _angle, color):
    """Wrapper for drawing triangle; called like `draw_triangle(ax, *coords, None, None, color)`"""
    ax.add_patch(
        Polygon(
            xy=[(x1, y1), (x2, y2), (x3, y3)], closed=True,
            facecolor=color, edgecolor=less_transparent_color(color)
        )
    )

def draw_text(ax, x, y, text, fontsize, color="black", **kwargs):
    """Wrapper for drawing text"""
    ax.text(
        x, y, text, fontsize=fontsize, color=color,
        horizontalalignment="center", verticalalignment="center", 
        zorder=10, **kwargs
    )

def generate_logics(n_sets):
    """Generate intersection identifiers in binary (0010 etc)"""
    for i in range(1, 2**n_sets):
        yield bin(i)[2:].zfill(n_sets)

def generate_petal_labels(datasets, fmt="{size}"):
    """Generate petal descriptions for venn diagram based on set sizes"""
    datasets = list(datasets)
    dataset_sizes = [len(dataset) for dataset in datasets]
    n_sets = len(datasets)
    dataset_union = set.union(*datasets)
    universe_size = len(dataset_union)
    petal_labels = {}
    for logic in generate_logics(n_sets):
        included_sets = [datasets[i] for i in range(n_sets) if logic[i] == "1"]
        excluded_sets = [datasets[i] for i in range(n_sets) if logic[i] == "0"]
        petal_set = (
            (dataset_union & set.intersection(*included_sets)) -
            set.union(set(), *excluded_sets)
        )
        if fmt is not None:
            petal_labels[logic] = fmt.format(
                logic=logic, size=len(petal_set),
                percentage=(100*len(petal_set)/universe_size)
            )
        else:
            petal_labels[logic] = len(petal_set)
    return petal_labels

def generate_bootstrap_pvalues(obs_labels, datasets, background):

    # overlaps are represented as binary strings, e.g. "110" for sets 1 and 2.

    def get_counts(petal_labels):
        counts = dict()
        for logic, size in petal_labels.items():
            size = int(size)
            # overlaps are counted only if they are not single sets. 
            if sum(map(int, logic)) == 1:
                continue
            # overlaps include nested overlaps ("111" counts are also
            # included in both the "110" counts). I.e. overlaps between all three stets
            # are also considered also overlaps between set one and two.
            for other_logic, other_size in petal_labels.items():
                other_size = int(other_size)            
                if logic == other_logic:
                    continue 

                if sum(map(int, logic)) == sum(i and j for i, j in zip(map(int, logic), map(int, other_logic))):
                    size += other_size
            counts[logic] = size

        return counts

    obs_counts = get_counts(obs_labels)

    datasets = list(datasets)
    dataset_sizes = [len(dataset) for dataset in datasets]

    # dataset_union = set.union(*datasets)
    # non_background = dataset_union.difference(set(background))
    # if non_background:
    #     print(f"Warning: {len(non_background)} genes not in background set are ignored", file=sys.stderr)

    # bootstrap
    boot_counts = defaultdict(list)
    nr_bootstraps = 10000
    for _ in range(nr_bootstraps):
        # new randomly sampled datasets of original sizes
        boot_datasets = []
        for s in dataset_sizes:
            boot_datasets.append(set(random.sample(background, s)))
        boot_labels = generate_petal_labels(boot_datasets)
        for logic, size in get_counts(boot_labels).items():
            boot_counts[logic].append(size)

    p_values = {}
    for logic, c in obs_counts.items():
        p_values[logic] = 0
        for bc in boot_counts[logic]:
            if bc >= c:
                p_values[logic] += 1/nr_bootstraps

    return p_values



# def generate_pvalues(datasets, maxp, background):
#     """Generate petal descriptions for venn diagram based on set sizes"""

#     from ..utils import fisher_test

#     datasets = list(datasets)
#     n_sets = len(datasets)
#     dataset_union = set.union(*datasets)
#     universe_size = len(dataset_union)
#     pvalues = {}
#     for logic in generate_logics(n_sets):
#         included_sets = [
#             datasets[i] for i in range(n_sets) if logic[i] == "1"
#         ]
#         if len(included_sets) == 2:
#             p = fisher_test(included_sets[0], included_sets[1], 
#                             background=background)
#             if p < maxp:
#                 pvalues[logic] = p

#     return pvalues

def init_axes(ax, figsize):
    """Create axes if do not exist, set axes parameters"""
    if ax is None:
        _, ax = subplots(nrows=1, ncols=1, figsize=figsize)
    ax.set(
        aspect="equal", frame_on=False,
        xlim=(-.05, 1.05), ylim=(-.05, 1.05),
        xticks=[], yticks=[]
    )
    return ax

def get_n_sets(petal_labels, dataset_labels):
    """Infer number of sets, check consistency"""
    n_sets = len(dataset_labels)
    for logic in petal_labels.keys():
        if len(logic) != n_sets:
            raise ValueError("Inconsistent petal and dataset labels")
        if not (set(logic) <= {"0", "1"}):
            raise KeyError("Key not understood: " + logic)
    return n_sets

def draw_venn(*, petal_labels, dataset_labels, pvalues, hint_hidden, colors, figsize, 
              fontsize, textcolor, shape_coords=None, legend_loc, ax):
    """Draw true Venn diagram, annotate petals and dataset labels"""
    n_sets = get_n_sets(petal_labels, dataset_labels)
    if 2 <= n_sets < 6:
        draw_shape = draw_ellipse
    elif n_sets == 6:
        draw_shape = draw_triangle
    else:
        raise ValueError("Number of sets must be between 2 and 6")
    ax = init_axes(ax, figsize)
    if shape_coords is None:
        shape_coords = SHAPE_COORDS[n_sets]
    shape_params = zip(
        shape_coords, SHAPE_DIMS[n_sets], SHAPE_ANGLES[n_sets], colors
        # SHAPE_COORDS[n_sets], SHAPE_DIMS[n_sets], SHAPE_ANGLES[n_sets], colors
    )
    for coords, dims, angle, color in shape_params:
        draw_shape(ax, *coords, *dims, angle, color)
    for logic, petal_label in petal_labels.items():
        # some petals could have been modified manually:
        if logic in PETAL_LABEL_COORDS[n_sets]:
            x, y = PETAL_LABEL_COORDS[n_sets][logic]


            # draw_text(ax, x, y, petal_label, fontsize=fontsize, color=textcolor)
            # if pvalues and logic in pvalues:
            #     p = max(pvalues[logic], 1/10000)
            #     stars = '\n' + '*' * (int(-np.log10(p/5)) - 1)                                       
            #     draw_text(ax, x, y, stars, fontsize=fontsize, 
            #                   color=textcolor, fontweight='bold'
            #                   ) 

                            
            if pvalues is not None:
                if logic in pvalues:
                    p = max(pvalues[logic], 1/10000)
                    stars = '*' * (int(-np.log10(p/5)) - 1)                                       

                    petal_label += '\n' + '*' * (int(-np.log10(p/5)) - 1)                                       
                    draw_text(ax, x, y, petal_label, fontsize=fontsize, 
                              color=textcolor, #fontweight='bold'
                              ) 
                else:                    
                    draw_text(ax, x, y, petal_label, fontsize=fontsize, 
                                 color=textcolor, 
                                #  color="#999999"
                              )                    
            else:
                draw_text(ax, x, y, petal_label, fontsize=fontsize, color=textcolor)

    if legend_loc is not None:
        ax.legend(dataset_labels, loc=legend_loc, prop={"size": fontsize})
    return ax

def update_hidden(hidden, logic, petal_labels):
    """Increment set's hidden count (sizes of intersections that are not displayed)"""
    for i, c in enumerate(logic):
        if c == "1":
            hidden[i] += int(petal_labels[logic])
    return hidden

def draw_hint_explanation(ax, dataset_labels, fontsize, textcolor):
    """Add explanation of 'n/d*' hints"""
    example_labels = list(dataset_labels)[0], list(dataset_labels)[3]
    hint_text = (
        "* elements of set in intersections that are not displayed,\n" +
        "such as shared only between {} and {}".format(*example_labels)
    )
    draw_text(ax, .5, -.1, hint_text, fontsize, textcolor)

def is_valid_dataset_dict(data):
    """Validate passed data (must be dictionary of sets)"""
    if not (hasattr(data, "keys") and hasattr(data, "values")):
        return False
    for dataset in data.values():
        if not isinstance(dataset, set):
            return False
    else:
        return True

def venn_dispatch(data, func, fmt="{size}", hint_hidden=False, cmap="Set2", 
                  alpha=.4, figsize=(8, 8), fontsize=13, textcolor='black', 
                  background=None, shape_coords=None, 
                  legend_loc="upper right", ax=None):
    """Check input, generate petal labels, draw venn diagram"""
    if not is_valid_dataset_dict(data):
        raise TypeError("Only dictionaries of sets are understood")
    n_sets = len(data)

    _data = data.copy()
    for k, v in data.items():
        _data[k] = _data[k].intersection(set(background))
        assert _data[k], f"Set {k} is empty after intersection with background"
        removed = len(data[k]) - len(_data[k])
        # print(f'Ignoring {removed} genes in "{k}" not part of background set', file=sys.stderr)
    data = _data

    petal_labels = generate_petal_labels(data.values(), fmt)

    pvalues = None
    if background:
        pvalues = generate_bootstrap_pvalues(petal_labels, data.values(), background)

        keys = list(pvalues.keys())
        for logic in keys:
            if pvalues[logic] > 0.05:
                del pvalues[logic]


    return func(
        petal_labels=petal_labels,
        dataset_labels=data.keys(), hint_hidden=hint_hidden,
        pvalues=pvalues,
        colors=generate_colors(n_colors=n_sets, cmap=cmap, alpha=alpha),
        figsize=figsize, fontsize=fontsize, textcolor=textcolor, 
        shape_coords=shape_coords, legend_loc=legend_loc, ax=ax,
    )

_venn = partial(venn_dispatch, func=draw_venn, hint_hidden=False)
    
def venn(*data, ncols=4, nrows=None, petals=3, palette='rainbow', 
          fontsize=9, textcolor=None, shape_coords=None, 
          figsize=None, alpha=0.4, background=None):

    import matplotlib.pyplot as plt
    import matplotlib
    from matplotlib.colors import ListedColormap

    if type(data[0]) is not tuple:
        _names = set()
        _data = []
        for x in data:
            assert x.name() not in _names, f"Duplicate name {x.name()} found in data"
            _names.add(x.name())
            _data.append((x.name(), set(x)))
        data = _data

    nrsets = len(data)

    petals = min(nrsets, petals)

    vals = np.linspace(0, 1, nrsets)
    viridis = matplotlib.colormaps[palette]
    combinations = list(itertools.combinations(range(nrsets), petals))

    ncols = min(ncols, len(combinations))

    if nrows is None:
        nrows = len(combinations)//ncols + int(len(combinations) % ncols > 0)

    if figsize is None:
        if nrsets == petals:
            figsize=(4, 3)
        else:
            figsize=(10, 2.5*nrows)

    fig, axes = plt.subplots(nrows, ncols, figsize=figsize, layout='constrained')

    if nrows == 1 and  ncols == 1:
        axes = np.array([axes])

    for i, ax in enumerate(axes.flat):
        if i >= len(combinations):
            ax.axis('off')
            continue
        combo = combinations[i]
        cmap = viridis([vals[i] for i in combo])
        cmap = ListedColormap(cmap)
        subset = dict([data[i] for i in combo])

        ax = _venn(subset, background=background, ax=ax, cmap=cmap, fontsize=fontsize, 
                   textcolor=textcolor, alpha=alpha, shape_coords=shape_coords)
        ax.get_legend().remove()
        ax.axis('off')

    handles = []
    cmap = viridis([vals[i] for i in range(len(data))])
    cmap = ListedColormap(cmap)
    for i, (label, _) in enumerate(data):
        cmap = viridis([vals[i]])
        cmap = ListedColormap(cmap)
        handles.extend([Patch(color=cmap(vals[i]), label=label, alpha=0.7)])

    if len(axes.flat) > len(combinations):
        l = fig.legend(loc='lower right', handles=handles, frameon=False, fontsize=fontsize, ncol=1)
    else:
        l = fig.legend(loc='outside center right', handles=handles, frameon=False, fontsize=fontsize, ncol=1)
    for text in l.get_texts():
        text.set_color(plt.rcParams['text.color'])


import unittest

import geneinfo.plot
import geneinfo.information
import geneinfo.utils
import geneinfo.genelist 
from geneinfo.genelist import GeneList

class TestGeneList(unittest.TestCase):

    def test_bitwise_or(self):
        self.assertEqual(
             GeneList(['A', 'B']) | GeneList(['C', 'B']),
             GeneList(['A', 'B', 'C']) 
            )
